########################################################################################################################
#                                          N-HANS speech denoiser: reader                                              #
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
#   Description:      Data reader pipeline                                            .                                #
#   Authors:          Shuo Liu, Gil Keren, Bjoern Schuller                                                             #
#   Affiliation:      Chair of Embedded Intelligence for Health Care and Wellbeing, University of Augsburg (UAU)  #
#   Version:          2.0                                                                                              #
#   Last Update:      May. 06, 2020                                                                                    #
#   Dependence Files: xxx                                                                                              #
#   Contact:          shuo.liu@informatik.uni-augburg.de                                                               #
########################################################################################################################
#                                                                                                                      #
#   Copyright (C) 2019, Shuo Liu, Gil Keren and Björn Schuller: University of Augsburg.                                #
#                                                                                                                      #
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public    #
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any       #
# later version.                                                                                                       #
#                                                                                                                      #
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied   #
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.#
#                                                                                                                      #
# You should have received a copy of the GNU General Public License along with this program. #  If not, see            #
# <http://www.gnu.org/licenses/>.                                                                                      #
########################################################################################################################

import numpy as np
import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
import tensorflow as tf
from tensorflow.python.util import deprecation
deprecation._PRINT_DEPRECATION_WARNINGS = False
tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
import random
import time
import hashlib
import pickle as p
from scipy.io.wavfile import read as wavread

"""
context_frames  : number of context frames          (reference signal)
window_frames   : number of frames of noisy signal  (input signal) 
random_slices   : number of slices (each slice contains one input signal and reference signals) 
eval_seeds      : 'valid' or 'test'. In training, evaluation is applied for 'valid' dataset, and in test, for 'test' dataset
wav_dump_folder : the folder to save denoised signals
speech_wav_dir  : the folder contains all speech .wav files
noise_wav_dir   : the folder contains all noise .wav files
"""

FLAGS = tf.compat.v1.flags.FLAGS
tf.compat.v1.flags.DEFINE_integer('context_frames', 200, '')
tf.compat.v1.flags.DEFINE_integer('window_frames', 35, '')
tf.compat.v1.flags.DEFINE_integer('random_slices', 50, '')
tf.compat.v1.flags.DEFINE_string('eval_seeds', 'valid', '')
tf.compat.v1.flags.DEFINE_string('wav_dump_folder', './wav_dump/', '')
tf.compat.v1.flags.DEFINE_string('speech_wav_dir', './speech_wav_dir/', '')
tf.compat.v1.flags.DEFINE_string('noise_wav_dir', './noise_wav_dir/', '')

"""
Create Pickle List for Training, Evaluation and Test
"""


# create speech pickle list
def create_speech_seeds():
    """
    Create 3 pickle files: train.pkl, valid.pkl and test.pkl for speech audio files
    :param speech_wav_dir: the directory contains three subfolders, train, valid and test, each includes audio .wav files
    """
    speech_wav_dir = FLAGS.speech_wav_dir
    train_folder_clean = speech_wav_dir + 'train'
    valid_folder_clean = speech_wav_dir + 'valid'
    test_folder_clean = speech_wav_dir + 'test'
    trainseeds = []
    validseeds = []
    testseeds = []
    wavfolders = [train_folder_clean, valid_folder_clean, test_folder_clean]
    seedsfolders = [trainseeds, validseeds, testseeds]

    for ii in range(3):
        wavfolder = wavfolders[ii]
        seedfolder = seedsfolders[ii]
        for root, directories, filenames in os.walk(wavfolder):
            for filename in filenames:
                if filename.endswith('.wav'):
                    seedfolder.append(os.path.join(root, filename))

    train_pkl = speech_wav_dir + 'train.pkl'
    valid_pkl = speech_wav_dir + 'valid.pkl'
    test_pkl = speech_wav_dir + 'test.pkl'
    p.dump(trainseeds, open(train_pkl, "wb"))
    p.dump(validseeds, open(valid_pkl, "wb"))
    p.dump(testseeds, open(test_pkl, "wb"))


# create noise pickle list
def create_noise_seeds():
    """
    Create 3 pickle files: train.pkl, valid.pkl and test.pkl for noise audio files
    :param noise_wav_dir: the directory contains three subfolders, train, valid and test, each includes audio .wav files
    """
    noise_wav_dir = FLAGS.noise_wav_dir
    train_folder_clean = noise_wav_dir + 'train'
    valid_folder_clean = noise_wav_dir + 'valid'
    test_folder_clean = noise_wav_dir + 'test'
    trainseeds = []
    validseeds = []
    testseeds = []
    wavfolders = [train_folder_clean, valid_folder_clean, test_folder_clean]
    seedsfolders = [trainseeds, validseeds, testseeds]

    for ii in range(3):
        wavfolder = wavfolders[ii]
        seedfolder = seedsfolders[ii]
        for root, directories, filenames in os.walk(wavfolder):
            for filename in filenames:
                if filename.endswith('.wav'):
                    seedfolder.append(os.path.join(root, filename))

    train_pkl = noise_wav_dir + 'train.pkl'
    valid_pkl = noise_wav_dir + 'valid.pkl'
    test_pkl = noise_wav_dir + 'test.pkl'
    p.dump(trainseeds, open(train_pkl, "wb"))
    p.dump(validseeds, open(valid_pkl, "wb"))
    p.dump(testseeds, open(test_pkl, "wb"))


""" 
Generating Audio Samples 
combine_signals: generating training, valid and test .wav files
"""


def read_wav(in_path):
    rate, samples = wavread(in_path)
    assert rate == 16000
    assert samples.dtype == 'int16'
    if len(samples.shape) > 1:
        samples = samples.mean(axis=1)
    assert len(samples.shape) == 1
    return samples


def domixing(cleansamples, noisepossamples, noisenegsamples, snr_pos, snr_neg):
    # Repeat the noise if it is shorter than the speech, or shorter it if it's longer
    nse_pos = noisepossamples
    nse_neg = noisenegsamples
    while len(cleansamples) - len(nse_pos) > 0:  # Make noise longer
        diff = len(cleansamples) - len(nse_pos)
        nse_pos = np.concatenate([nse_pos, noisepossamples[:diff]], axis=0)

    while len(cleansamples) - len(nse_neg) > 0:  # Make noise longer
        diff = len(cleansamples) - len(nse_neg)
        nse_neg = np.concatenate([nse_neg, noisenegsamples[:diff]], axis=0)

    if len(cleansamples) - len(noisepossamples) < 0:  # Make noise shorter
        nse_pos = noisepossamples[:len(cleansamples)]

    if len(cleansamples) - len(noisenegsamples) < 0:  # Make noise shorter
        nse_neg = noisenegsamples[:len(cleansamples)]

    sig = cleansamples

    # Power of signal and noise
    psignal = sum(abs(sig) * abs(sig)) / sig.shape[0]
    pnoise_pos = sum(abs(nse_pos) * abs(nse_pos)) / nse_pos.shape[0]
    pnoise_neg = sum(abs(nse_neg) * abs(nse_neg)) / nse_neg.shape[0]

    # Compute scale factor
    if pnoise_pos == 0:
        K_pos = 1
    else:
        K_pos = (psignal / pnoise_pos) * pow(10, -snr_pos / 10.0)
        K_pos = np.sqrt(K_pos)

    if pnoise_neg == 0:
        K_neg = 1
    else:
        K_neg = (psignal / pnoise_neg) * pow(10, -snr_neg / 10.0)
        K_neg = np.sqrt(K_neg)

    # Mix
    noise_pos_scaled = K_pos * nse_pos  # Scale the noise
    noise_neg_scaled = K_neg * nse_neg  # Scale the noise
    mixed = sig + noise_pos_scaled + noise_neg_scaled  # Mix
    mixed = mixed / (max(abs(mixed)) + 0.000001)  # Normalize
    target = sig + noise_pos_scaled
    target = target / (max(abs(mixed)) + 0.000001)  # Normalize
    noise_pos_signal = noise_pos_scaled / (max(abs(mixed)) + 0.000001)
    noise_neg_signal = noise_neg_scaled / (max(abs(mixed)) + 0.000001)

    return mixed, target, K_pos, K_neg, noise_pos_signal, noise_neg_signal


def combine_signals(istrain, cleanpath, noisepospath, noisenegpath):
    try:
        # Read Wavs
        cleansamples = read_wav(cleanpath)
        noisepossamples = read_wav(noisepospath)
        noisenegsamples = read_wav(noisenegpath)

        # Normalize
        cleansamples = cleansamples / (max(abs(cleansamples)) + 0.000001)
        noisepossamples = noisepossamples / (max(abs(noisepossamples)) + 0.000001)
        noisenegsamples = noisenegsamples / (max(abs(noisenegsamples)) + 0.000001)
        cleansamples = cleansamples.astype(np.float32)
        noisepossamples = noisepossamples.astype(np.float32)
        noisenegsamples = noisenegsamples.astype(np.float32)

        # Cut the end to have an exact number of frames
        if (len(cleansamples) - 400) % 160 != 0:
            cleansamples = cleansamples[:-((len(cleansamples) - 400) % 160)]

        # Choose SNR
        SNRs = [-3, 0, 3, 5, 8]

        if istrain:
            snr_pos = SNRs[random.randint(0, len(SNRs) - 1)]
            snr_neg = SNRs[random.randint(0, len(SNRs) - 1)]
            mixed, target, Kpos, Kneg, noisepossignal, noisenegsignal = domixing(cleansamples, noisepossamples,
                                                                                 noisenegsamples, snr_pos, snr_neg)
            return target, noisepossignal, noisenegsignal, mixed, np.array(snr_pos, dtype=np.int32), np.array(snr_neg,
                                                                                                              dtype=np.int32)

        else:
            # The noise is the beginning of the file, and mix the rest. The snrs for validation and test data depend on data names, and hence consistent during validation and test
            snrid_pos = int(hashlib.md5(cleanpath).hexdigest()[:8], 16) % len(SNRs)
            snrid_neg = int(hashlib.md5(cleanpath).hexdigest()[:6], 16) % len(SNRs)
            snr_pos = SNRs[snrid_pos]
            snr_neg = SNRs[snrid_neg]
            mixed, target, Kpos, Kneg, noisepossignal, noisenegsignal = domixing(cleansamples, noisepossamples,
                                                                                 noisenegsamples, snr_pos, snr_neg)
            return target, noisepossignal, noisenegsignal, mixed, np.array(snr_pos, dtype=np.int32), np.array(snr_neg,
                                                                                                              dtype=np.int32)

    except:
        print('error in threads')
        print(cleanpath, noisepospath, noisenegpath)


######################################     Global FLAGS and Training FLAGS     #########################################
class read_seeds:
    """
    name              : 'train', 'valid' or 'test'. To generate data reader.
    queuesize         : the size of data queue
    min_after_dequeue : minimum seeds left in queue after dequeue operation
    nthreads          : number of threads processing dequeue
    """

    def __init__(self, name, queuesize, min_after_dequeue, nthreads):
        self.Fs = 16000
        self.frame_length = int(self.Fs * 0.025)
        self.frame_step = int(self.Fs * 0.010)
        self.queuesize = queuesize
        self.min_after_dequeue = min_after_dequeue
        self.window_frames = FLAGS.window_frames
        self.random_slices = FLAGS.random_slices
        self.context_frames = FLAGS.context_frames
        self.eval_stride = 1
        self.istrain = name == 'train'
        self.nthreads = nthreads
        self.name = name

        if self.istrain != True:
            name = FLAGS.eval_seeds

        speech_wav_dir = FLAGS.speech_wav_dir
        noise_wav_dir = FLAGS.noise_wav_dir
        if name == 'train':
            self.seedspaths = [speech_wav_dir + 'train.pkl', noise_wav_dir + 'train.pkl']
        elif name == 'valid':
            self.seedspaths = [speech_wav_dir + 'valid.pkl', noise_wav_dir + 'valid.pkl']
        elif name == 'test':
            self.seedspaths = [speech_wav_dir + 'test.pkl', noise_wav_dir + 'test.pkl']

    def preparations(self):
        """
        Preparation of the data queue.
        """
        self.seeds = []
        for seedspath in self.seedspaths:
            if seedspath.endswith('.pkl'):
                with open(seedspath, 'rb') as f:
                    seedlist = p.load(f)
                    seedlist = tf.constant(np.array(seedlist))
            self.seeds.append(seedlist)

        # Construct seeds queues
        self.seedsqs = []
        for seed in self.seeds:
            num_epochs = None if self.istrain else 1
            seedsq = tf.train.input_producer(seed, capacity=50000, shuffle=self.istrain, num_epochs=num_epochs)
            self.epochcounter = None if self.istrain else tf.compat.v1.local_variables()[-1]
            self.seedsqs.append(seedsq)

        # Get the examples from seeds. It is a list of tensors, with first dimension being the example.
        examples = self.get_examples(self.istrain, self.seedsqs)

        # Create examples queue and enqueue the examples
        dtypes = [x.dtype for x in examples]
        shapes = [x.shape.as_list()[1:] for x in examples]
        if self.istrain:
            exq = tf.queue.RandomShuffleQueue(capacity=self.queuesize, min_after_dequeue=self.min_after_dequeue,
                                              dtypes=dtypes, shapes=shapes)
        else:
            exq = tf.queue.FIFOQueue(capacity=self.queuesize, dtypes=dtypes, shapes=shapes)
        enqop = exq.enqueue_many(examples)

        # The dequeue operation and the mb placeholder
        mb = tf.compat.v1.placeholder(dtype=tf.int32, shape=[], name=self.name + '_mb')
        if self.istrain:
            batch = exq.dequeue_many(mb)
        else:
            batch = exq.dequeue_up_to(mb)

        # Create queue runner
        qr = tf.train.QueueRunner(exq, [enqop] * self.nthreads)
        tf.train.add_queue_runner(qr, tf.compat.v1.GraphKeys.QUEUE_RUNNERS)

        # Set some members
        self.exq = exq
        self.batch = batch
        self.mb = mb

    def get_examples(self, istrain, seedsqs):
        """
        get model inputs from audio samples
        :param istrain: boolean value. If True, the examples are randomly trauncated from audio samples; If False, the examples are trauncated frame by frame from audio samples
        :param seedsqs: seed lists of speech and noise .wav files
        :return:
        """
        clean_seed = seedsqs[0].dequeue()
        noise_pos_seed = seedsqs[1].dequeue()
        noise_neg_seed = seedsqs[1].dequeue()

        target_wav, noise_pos_wav, noise_neg_wav, mix_wav, snr_pos, snr_neg = tf.py_func(combine_signals,
                                                                                         [istrain, clean_seed,
                                                                                          noise_pos_seed,
                                                                                          noise_neg_seed],
                                                                                         [tf.float32, tf.float32,
                                                                                          tf.float32, tf.float32,
                                                                                          tf.int32, tf.int32])

        # flatten
        target_wav, noise_pos_wav, noise_neg_wav, mix_wav = [tf.reshape(x, [-1]) for x in
                                                             (target_wav, noise_pos_wav, noise_neg_wav, mix_wav)]

        # stft
        target_fft = tf.signal.stft(target_wav, self.frame_length, self.frame_step, fft_length=self.frame_length)
        noise_pos_fft = tf.signal.stft(noise_pos_wav, self.frame_length, self.frame_step, fft_length=self.frame_length)
        noise_neg_fft = tf.signal.stft(noise_neg_wav, self.frame_length, self.frame_step, fft_length=self.frame_length)
        mix_fft = tf.signal.stft(mix_wav, self.frame_length, self.frame_step, fft_length=self.frame_length)
        # Magnitude
        target_mag, noise_pos_mag, noise_neg_mag, mix_mag = [tf.abs(x) for x in
                                                             [target_fft, noise_pos_fft, noise_neg_fft, mix_fft]]
        # Phase
        # temp_wav = mix_wav
        mix_phase = tf.math.angle(mix_fft)
        target_phase = tf.math.angle(target_fft)
        pos_phase = tf.math.angle(noise_pos_fft)
        neg_phase = tf.math.angle(noise_neg_fft)

        # log
        target_log, noise_pos_log, noise_neg_log, mix_log = [tf.math.log(x + 1e-5) for x in
                                                             [target_mag, noise_pos_mag, noise_neg_mag, mix_mag]]

        if istrain:
            # padding
            target_log, noise_pos_log, noise_neg_log, mix_log, mix_phase, target_phase, pos_phase, neg_phase = [
                self.pad_1D_for_windowing(x, self.window_frames) for x in
                [target_log, noise_pos_log, noise_neg_log, mix_log, mix_phase, target_phase, pos_phase, neg_phase]]

            # Get n random window-label pairs
            targetslices = []
            targetphslices = []
            mixedslices = []
            mixedphslices = []
            posslices = []
            negslices = []
            posphslices = []
            negphslices = []

            noise_neg_slices = []
            noise_pos_slices = []

            for _ in range(self.random_slices):
                # Crop a window
                targetcropped, mixedcropped, mixedphcropped, targetphcropped, poscropped, posphcropped, negcropped, negphcropped, noiseposcontext, noisenegcontext = \
                    self.synchronized_1D_crop(target_log, noise_pos_log, noise_neg_log, mix_log, mix_phase,
                                              target_phase, pos_phase, neg_phase, self.window_frames,
                                              self.context_frames)

                # Add to list
                for (lst, tnsr) in zip(
                        [targetslices, mixedslices, mixedphslices, targetphslices, posslices, posphslices, negslices,
                         negphslices, noise_pos_slices, noise_neg_slices],
                        [targetcropped, mixedcropped, mixedphcropped, targetphcropped, poscropped, posphcropped,
                         negcropped, negphcropped, noiseposcontext, noisenegcontext]):
                    lst.append(tnsr)

            # Stack and return
            return tf.stack(targetslices, axis=0), tf.stack(mixedslices, axis=0), tf.stack(mixedphslices,
                                                                                           axis=0), tf.stack(
                targetphslices, axis=0), tf.stack(posslices, axis=0), tf.stack(posphslices, axis=0), tf.stack(negslices,
                                                                                                              axis=0), tf.stack(
                negphslices, axis=0), \
                   tf.stack(noise_pos_slices, axis=0), tf.stack(noise_neg_slices, axis=0), \
                   tf.constant([0] * self.random_slices, dtype=tf.int32), tf.constant([''] * self.random_slices,
                                                                                      dtype=tf.string), \
                   tf.constant([''] * self.random_slices, dtype=tf.string), tf.constant([''] * self.random_slices,
                                                                                        dtype=tf.string), tf.constant(
                [0] * self.random_slices, dtype=tf.int32), tf.constant([0] * self.random_slices, dtype=tf.int32)
        else:
            targetslices = self.strided_crop(target_log[self.context_frames:], 1, self.eval_stride)
            mixedslices = self.strided_crop(mix_log[self.context_frames:], self.window_frames, self.eval_stride)
            mixedphslices = self.strided_crop(mix_phase[self.context_frames:], 1, self.eval_stride)
            targetphslices = self.strided_crop(target_phase[self.context_frames:], 1, self.eval_stride)
            posslices = self.strided_crop(noise_pos_log[self.context_frames:], self.window_frames, self.eval_stride)
            posphslices = self.strided_crop(pos_phase[self.context_frames:], 1, self.eval_stride)
            negslices = self.strided_crop(noise_neg_log[self.context_frames:], self.window_frames, self.eval_stride)
            negphslices = self.strided_crop(neg_phase[self.context_frames:], 1, self.eval_stride)

            noiseposcontext = noise_pos_log[:self.context_frames]
            noiseposcontext = tf.reshape(noiseposcontext, [self.context_frames, noiseposcontext.shape[1].value])
            noise_pos_slices = tf.tile(tf.expand_dims(noiseposcontext, 0), [tf.shape(targetslices)[0], 1, 1])
            noisenegcontext = noise_neg_log[:self.context_frames]
            noisenegcontext = tf.reshape(noisenegcontext, [self.context_frames, noisenegcontext.shape[1].value])
            noise_neg_slices = tf.tile(tf.expand_dims(noisenegcontext, 0), [tf.shape(targetslices)[0], 1, 1])
            locations = tf.range(0, tf.shape(targetslices)[0], dtype=tf.int32)
            clean_seeds = tf.tile([clean_seed], [tf.shape(targetslices)[0]])
            noise_pos_seeds = tf.tile([noise_pos_seed], [tf.shape(targetslices)[0]])
            noise_neg_seeds = tf.tile([noise_neg_seed], [tf.shape(targetslices)[0]])
            snrs_pos = tf.tile([snr_pos], [tf.shape(targetslices)[0]])
            snrs_neg = tf.tile([snr_neg], [tf.shape(targetslices)[0]])
            return targetslices, mixedslices, mixedphslices, targetphslices, posslices, posphslices, negslices, negphslices, noise_pos_slices, noise_neg_slices, locations, clean_seeds, noise_pos_seeds, noise_neg_seeds, snrs_pos, snrs_neg  # , temp_wav

    def pad_1D_for_windowing(self, tensor, length):
        """
        Padding the input spectrogramm
        :param tensor: shape of [Times x Frequencies]
        :param length: equals to window frames
        :return: padded spectrogram
        """
        len_before = ((length + 1) // 2) - 1
        len_after = length // 2
        return tf.pad(tensor, [[len_before, len_after], [0, 0]])

    def synchronized_1D_crop(self, target, noise_pos, noise_neg, mixed, mixedph, targetph, posph, negph, winlength,
                             contextlength):
        winmaxval = tf.shape(mixed)[0] - winlength
        winstart = tf.random.uniform([], dtype=tf.int32, minval=0, maxval=winmaxval + 1)

        # Crop window
        mixedcropped = mixed[winstart:winstart + winlength]
        targetcropped = target[winstart:winstart + winlength]
        mixedphcropped = mixedph[winstart:winstart + winlength]
        mixedcropped = tf.reshape(mixedcropped, [winlength, mixedcropped.shape[1].value])
        targetcropped = tf.reshape(targetcropped, [winlength, targetcropped.shape[1].value])
        mixedphcropped = tf.reshape(mixedphcropped, [winlength, mixedphcropped.shape[1].value])

        # Get the center of the clean and mixed phase
        targetcropped = targetcropped[self.window_frames // 2]
        targetcropped = tf.expand_dims(targetcropped, 0)
        mixedphcropped = mixedphcropped[self.window_frames // 2]
        mixedphcropped = tf.expand_dims(mixedphcropped, 0)

        # log.....
        poscropped = noise_pos[winstart:winstart + winlength]
        poscropped = tf.reshape(poscropped, [winlength, poscropped.shape[1].value])
        negcropped = noise_neg[winstart:winstart + winlength]
        negcropped = tf.reshape(negcropped, [winlength, negcropped.shape[1].value])

        # ph.....
        targetphcropped = targetph[winstart:winstart + winlength]
        targetphcropped = tf.reshape(targetphcropped, [winlength, targetphcropped.shape[1].value])
        targetphcropped = targetphcropped[self.window_frames // 2]
        targetphcropped = tf.expand_dims(targetphcropped, 0)

        negphcropped = negph[winstart:winstart + winlength]
        negphcropped = tf.reshape(negphcropped, [winlength, negphcropped.shape[1].value])
        negphcropped = negphcropped[self.window_frames // 2]
        negphcropped = tf.expand_dims(negphcropped, 0)

        posphcropped = posph[winstart:winstart + winlength]
        posphcropped = tf.reshape(posphcropped, [winlength, posphcropped.shape[1].value])
        posphcropped = posphcropped[self.window_frames // 2]
        posphcropped = tf.expand_dims(posphcropped, 0)

        # The rest of the noise
        noiseposrest = tf.concat([noise_pos[:winstart], noise_pos[winstart + winlength:]], axis=0)
        noisenegrest = tf.concat([noise_neg[:winstart], noise_neg[winstart + winlength:]], axis=0)

        # Get the noise part
        posrestmaxval = tf.shape(noiseposrest)[0] - contextlength
        posreststart = tf.random.uniform([], dtype=tf.int32, minval=0, maxval=posrestmaxval + 1)
        noiseposcontext = noiseposrest[posreststart:posreststart + contextlength]
        noiseposcontext = tf.reshape(noiseposcontext, [contextlength, noiseposcontext.shape[1].value])

        negrestmaxval = tf.shape(noisenegrest)[0] - contextlength
        negreststart = tf.random.uniform([], dtype=tf.int32, minval=0, maxval=negrestmaxval + 1)
        noisenegcontext = noisenegrest[negreststart:negreststart + contextlength]
        noisenegcontext = tf.reshape(noisenegcontext, [contextlength, noisenegcontext.shape[1].value])

        return targetcropped, mixedcropped, mixedphcropped, targetphcropped, poscropped, posphcropped, negcropped, negphcropped, noiseposcontext, noisenegcontext

    def strided_crop(self, tensor, length, stride):
        # we assume that we have a length dimension and a feature dimension
        assert len(tensor.shape) == 2
        n_features = int(tensor.shape[1])
        padded = self.pad_1D_for_windowing(tensor, length)
        windows = tf.extract_image_patches(tf.expand_dims(tf.expand_dims(padded, axis=0), axis=3),
                                           ksizes=[1, length, n_features, 1],
                                           strides=[1, stride, n_features, 1],
                                           rates=[1, 1, 1, 1],
                                           padding='VALID')
        return tf.reshape(windows, [-1, length, n_features])

    def get_inputs(self):
        if not hasattr(self, 'ph'):
            self.ph = [tf.compat.v1.placeholder(dtype=tf.float32, shape=[None, 1, 201], name='targetph'),
                       tf.compat.v1.placeholder(dtype=tf.float32, shape=[None, self.window_frames, 201],
                                                name='mixedph'),
                       tf.compat.v1.placeholder(dtype=tf.float32, shape=[None, 1, 201], name='mixedphaseph'),
                       tf.compat.v1.placeholder(dtype=tf.float32, shape=[None, 1, 201], name='targetphaseph'),
                       tf.compat.v1.placeholder(dtype=tf.float32, shape=[None, self.window_frames, 201], name='posph'),
                       tf.compat.v1.placeholder(dtype=tf.float32, shape=[None, 1, 201], name='posphaseph'),
                       tf.compat.v1.placeholder(dtype=tf.float32, shape=[None, self.window_frames, 201], name='negph'),
                       tf.compat.v1.placeholder(dtype=tf.float32, shape=[None, 1, 201], name='negphaseph'),
                       tf.compat.v1.placeholder(dtype=tf.float32, shape=[None, self.context_frames, 201],
                                                name='noiseposcontextph'),
                       tf.compat.v1.placeholder(dtype=tf.float32, shape=[None, self.context_frames, 201],
                                                name='noisenegcontextph'),
                       tf.compat.v1.placeholder(dtype=tf.int32, shape=[None], name='locationph'),
                       tf.compat.v1.placeholder(dtype=tf.string, shape=[None], name='targetpathph'),
                       tf.compat.v1.placeholder(dtype=tf.string, shape=[None], name='noisepospathph'),
                       tf.compat.v1.placeholder(dtype=tf.string, shape=[None], name='noisenegpathph'),
                       tf.compat.v1.placeholder(dtype=tf.int32, shape=[None], name='snrposph'),
                       tf.compat.v1.placeholder(dtype=tf.int32, shape=[None], name='snrnegph'),
                       ]
        return self.ph


if __name__ == '__main__':
    g = tf.Graph()
    with g.as_default():
        reader = read_seeds(name='valid', queuesize=200, min_after_dequeue=100, nthreads=16)
        with tf.device('/cpu:0'):
            reader.preparations()

        print('starting queue runners')
        sess = tf.compat.v1.Session(graph=g, config=tf.compat.v1.ConfigProto(allow_soft_placement=True))
        sess.run(tf.compat.v1.local_variables_initializer())
        coord = tf.train.Coordinator()
        threads = tf.train.start_queue_runners(sess=sess, coord=coord)
        queue_stats = tf.stack(
            [x.queue.size() for x in tf.compat.v1.get_collection(tf.compat.v1.GraphKeys.QUEUE_RUNNERS)])
        feed_dict = {reader.mb: 100}

        with tf.device('/gpu:0'):
            target, mixed, mixedph, targetph, pos, posph, neg, negph, noiseposcontext, noisenegcontext, location, cleanpath, noisepospath, noisenegpath, snr_pos, snr_neg = reader.get_inputs()

        for i in range(3):
            print(i)
            t1 = time.time()
            target_, mixed_, mixedph_, targetph_, pos_, posph_, neg_, negph_, noiseposcontext_, noisenegcontext_, location_, cleanpath_, noisepospath_, noisenegpath_, snr_pos_, snr_neg_ = sess.run(
                reader.batch, feed_dict=feed_dict)
            t2 = time.time()
            print(sess.run(queue_stats), t2 - t1)




########################################################################################################################
#                                          N-HANS speech separator: apply                                              #
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
#   Description:      Apply N-HANS trained separator for audio.                                                        #
#   Authors:          Shuo Liu, Gil Keren, Bjoern Schuller                                                             #
#   Affiliation:      Chair of Embedded Intelligence for Health Care and Wellbeing, University of Augsburg (UAU)  #
#   Version:          2.0                                                                                              #
#   Last Update:      May. 06, 2020                                                                                    #
#   Dependence Files: main.py                                                                                          #
#   Contact:          shuo.liu@informatik.uni-augburg.de                                                               #
########################################################################################################################
#                                                                                                                      #
#   Copyright (C) 2019, Shuo Liu, Gil Keren and Björn Schuller: University of Augsburg.                                #
#                                                                                                                      #
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public    #
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any       #
# later version.                                                                                                       #
#                                                                                                                      #
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied   #
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.#
#                                                                                                                      #
# You should have received a copy of the GNU General Public License along with this program. #  If not, see            #
# <http://www.gnu.org/licenses/>.                                                                                      #
#                                                                                                                      #
########################################################################################################################

from __future__ import division, absolute_import, print_function
import os
import sox
import math, functools
import numpy as np
import warnings
warnings.filterwarnings('ignore', category=FutureWarning)
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
import tensorflow as tf
if int(tf.__version__.split('.')[0]) == 2:
    import tensorflow.compat.v1 as tf
    tf.compat.v1.disable_v2_behavior()
from tensorflow.python.util import deprecation
deprecation._PRINT_DEPRECATION_WARNINGS = False
tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
from scipy.io.wavfile import read as wavread
from scipy.io.wavfile import write as wavwrite
# from main import model
#            ||
#            \/
from N_HANS.N_HANS___Source_Separation import main
model = main.model

FLAGS = tf.compat.v1.flags.FLAGS

tf.compat.v1.flags.DEFINE_string('trained_model_dir', './trained_model', '')
tf.compat.v1.flags.DEFINE_string('trained_model_name', '81457_2-545000', '')

tf.compat.v1.flags.DEFINE_string('input', './audio_examples/mixed.wav', '')
tf.compat.v1.flags.DEFINE_string('neg', './audio_examples/noise_speaker.wav', '')
tf.compat.v1.flags.DEFINE_string('pos', './audio_examples/target_speaker.wav', '')
tf.compat.v1.flags.DEFINE_string('output', './audio_examples/denoised.wav', '')

tf.compat.v1.flags.DEFINE_float('compensate', 0, '')
tf.compat.v1.flags.DEFINE_boolean('ac', False, '')

Noise_Win = 200
Mix_Win = 35


"""
Assistence Functions
"""


def format_input_audio(mixed, pos, neg, out):
    input_audio = [mixed, pos, neg]
    input_names = ['mixed', 'pos', 'neg']
    fs = int(sox.file_info.sample_rate(mixed))
    br = int(sox.file_info.bitrate(mixed))
    # check existence
    for idx, audio_file in enumerate([mixed, pos, neg]):
        if not os.path.exists(audio_file):
            print('+++++++++++++++++++++++++++++++++')
            print('Please check. {} path not exists.'.format(input_names[idx]))

    # converter
    soxer = sox.Transformer()
    soxer.convert(samplerate=16000, n_channels=1, bitdepth=16)

    for idx, audio_file in enumerate([mixed, pos, neg]):
        format_satisfied = audio_file.endswith('.wav')
        samplerate_satisfied = int(sox.file_info.sample_rate(audio_file)) == 16000
        satisfied = format_satisfied * samplerate_satisfied
        if format_satisfied:
            bitrate_satisfied = int(sox.file_info.bitrate(audio_file)) == 16
            satisfied = format_satisfied * samplerate_satisfied * bitrate_satisfied
        if not satisfied:
            if not os.path.exists('./temp'):
                os.mkdir('./temp')
            temp_file = './temp/' + input_names[idx] + '_temp.wav'
            soxer.build(audio_file, temp_file)
            input_audio[idx] = temp_file

    out_format = sox.file_info.file_extension(out)
    if out_format != 'wav':
        out = out.replace(out_format, 'wav')
    return input_audio, out, fs, br


def format_output_audio(out_file, fs, br):
    format = sox.file_info.file_extension(out_file)
    if format != 'wav':
        soxer = sox.Transformer()
        soxer.convert(samplerate=fs, bitdepth=br)
        soxer.build(out_file.replace(format, 'wav'), out_file)
        os.remove(out_file.replace(format, 'wav'))

def read_wav(in_path):
    rate, samples = wavread(in_path)
    assert rate == 16000
    assert samples.dtype == 'int16'
    if len(samples.shape) > 1:
        samples = samples.mean(axis=1)
    assert len(samples.shape) == 1
    return samples


def domixing(cleansamples, noisesamples, snr):
    # Repeat the noise if it is shorter than the speech, or shorter it if it's longer
    nse = noisesamples
    while len(cleansamples) - len(nse) > 0: # Make noise longer
        diff = len(cleansamples) - len(nse)
        nse = np.concatenate([nse, noisesamples[:diff]], axis=0)
    if len(cleansamples) - len(noisesamples) < 0: # Make noise shorter
        nse = noisesamples[:len(cleansamples)]
    sig = cleansamples

    # Power of signal and noise
    psignal = sum(abs(sig) * abs(sig)) / sig.shape[0]
    pnoise = sum(abs(nse) * abs(nse)) / nse.shape[0]

    # Compute scale factor
    if pnoise == 0:
        K = 1
    else:
        K = (psignal / pnoise) * pow(10, -snr / 10.0)
    K = np.sqrt(K)

    # Mix
    noise_scaled = K * nse                                # Scale the noise
    mixed = sig + noise_scaled                            # Mix
    mixed = mixed / (max(abs(mixed))+0.000001)            # Normalize
    return mixed, K


def combine_signals(cleanpath, noisepath):
    try:
        # Read Wavs
        cleansamples = read_wav(cleanpath)
        noisesamples = read_wav(noisepath)

        # Normalize
        cleansamples = cleansamples / (max(abs(cleansamples))+0.000001)
        noisesamples = noisesamples / (max(abs(noisesamples))+0.000001)
        cleansamples = cleansamples.astype(np.float32)
        noisesamples = noisesamples.astype(np.float32)

        # Cut the end to have an exact number of frames
        cleansamples = cleansamples[:-((len(cleansamples) - 400) % 160)]

        # Choose SNR
        SNRs = [-5, -3, -1, 0, 1, 3, 5]
        # The noise is the beginning of the file, and mix the rest
        snr = 0
        mixed, K = domixing(cleansamples, noisesamples, snr)
        return cleansamples, noisesamples * K, mixed, np.array(snr, dtype=np.int32)
    except:
        print('error in threads')
        print(cleanpath, noisepath)


def handle_signals(mixedpath, cleanpath, noisepath):
    try:
        # Read Wavs
        mixedsamples = read_wav(mixedpath)
        cleansamples = read_wav(cleanpath)
        noisesamples = read_wav(noisepath)

        # Normalize
        mixedsamples = mixedsamples / (max(abs(mixedsamples))+0.000001)
        cleansamples = cleansamples / (max(abs(cleansamples))+0.000001)
        noisesamples = noisesamples / (max(abs(noisesamples))+0.000001)
        mixedsamples = mixedsamples.astype(np.float32)
        cleansamples = cleansamples.astype(np.float32)
        noisesamples = noisesamples.astype(np.float32)

        desired_len = int(2.015 * 16000)
        while desired_len - len(cleansamples) > 0:  # Make noise longer
            diff = desired_len - len(cleansamples)
            cleansamples = np.concatenate([cleansamples, cleansamples[:diff]], axis=0)

        while desired_len - len(noisesamples) > 0:  # Make noise longer
            diff = desired_len - len(noisesamples)
            noisesamples = np.concatenate([noisesamples, noisesamples[:diff]], axis=0)

        # Cut the end to have an exact number of frames
        if (len(mixedsamples) - 400) % 160 != 0:
            mixedsamples = mixedsamples[:-((len(mixedsamples) - 400) % 160)]

        return cleansamples, noisesamples, mixedsamples

    except:
        print('error in threads')
        print(mixedpath, cleanpath, noisepath)


def pad_1D_for_windowing(tensor, length):
    len_before = ((length + 1) // 2) - 1
    len_after = length // 2
    return tf.pad(tensor, [[len_before, len_after],[0,0]])


def strided_crop(tensor, length, stride):
    # we assume that we have a length dimension and a feature dimension
    assert len(tensor.shape) == 2
    n_features = int(tensor.shape[1])
    padded = pad_1D_for_windowing(tensor, length)
    windows = tf.extract_image_patches(tf.expand_dims(tf.expand_dims(padded, axis=0), axis=3),
                                           ksizes=[1, length, n_features, 1],
                                           strides=[1, stride, n_features, 1],
                                           rates=[1, 1, 1, 1],
                                           padding='VALID')
    return tf.reshape(windows, [-1, length, n_features])


def recover_samples_from_spectrum(logspectrum_stft, spectrum_phase, save_to=None):
    abs_spectrum = np.exp(logspectrum_stft)
    spectrum = abs_spectrum * (np.exp(1j * spectrum_phase))
    istft_graph = tf.Graph()
    with istft_graph.as_default():
        stft_ph = tf.placeholder(tf.complex64, shape=(None, 201))
        samples = tf.signal.inverse_stft(stft_ph, 400, 160, 400, window_fn=tf.signal.inverse_stft_window_fn(160, forward_window_fn=functools.partial(tf.signal.hann_window, periodic=True)))
        istft_sess = tf.Session(config=tf.ConfigProto(allow_soft_placement=True))
        samples_ = istft_sess.run(samples, feed_dict={stft_ph: spectrum})
        if save_to is not None:
            wavwrite(save_to, 16000, samples_)


"""
Apply N-HANS speech separater
"""


def apply_demo(cleanpath, noisepath, save_to):
    cleanlist = []
    cleanlist.append(cleanpath)
    noiselist = []
    noiselist.append(noisepath)

    # data processing
    g = tf.Graph()
    with g.as_default():
        with tf.device('/cpu:0'):
            cleanlist = tf.constant(np.array(cleanlist))
            noiselist = tf.constant(np.array(noiselist))

            cleanpath_ph = cleanlist[0]
            noisepath_ph = noiselist[0]

            clean_wav, noise_wav, mix_wav, snr = tf.py_func(combine_signals,
                                                            [cleanpath_ph, noisepath_ph],
                                                            [tf.float32, tf.float32, tf.float32, tf.int32])

            clean_wav, noise_wav, mix_wav = [tf.reshape(x, [-1]) for x in (clean_wav, noise_wav, mix_wav)]

            sess = tf.Session(config=tf.ConfigProto(allow_soft_placement=True))

            mix_stft, clean_stft, noise_stft = [tf.signal.stft(wav, 400, 160, fft_length=400) for wav in [mix_wav, clean_wav, noise_wav]]

            mix_spectrum, clean_spectrum, noise_spectrum = [tf.log(tf.abs(wav_stft) + 1e-5) for wav_stft in [mix_stft, clean_stft, noise_stft]]

            mix_phase = tf.angle(mix_stft)


            # crop data
            mix_spectra = strided_crop(mix_spectrum[Noise_Win:], Mix_Win, 1)

            # postive noise & negative noise
            clean_spectrum = clean_spectrum[:Noise_Win]
            clean_spectrum = tf.reshape(clean_spectrum, [Noise_Win, clean_spectrum.shape[1].value])
            clean_spectra = tf.tile(tf.expand_dims(clean_spectrum, 0), [tf.shape(mix_spectra)[0], 1, 1])

            noise_spectrum = noise_spectrum[:Noise_Win]
            noise_spectrum = tf.reshape(noise_spectrum, [Noise_Win, noise_spectrum.shape[1].value])
            noise_spectra = tf.tile(tf.expand_dims(noise_spectrum, 0), [tf.shape(mix_spectra)[0], 1, 1])

            mixedphs = strided_crop(mix_phase[Noise_Win:], 1, 1)

            # get data
            print('--------------------------------')
            mix_spectra_, clean_spectra_, noise_spectra_ = sess.run([mix_spectra, clean_spectra, noise_spectra])
            mixedphs_ = sess.run(mixedphs, feed_dict={cleanpath_ph: cleanpath,
                                                      noisepath_ph: noisepath})

    mb = 100
    batches = int(math.ceil(len(mix_spectra_) / float(mb)))
    # batches = 1
    denoised = []
    mix_centers = []

    # denoising
    # graph
    eg = tf.Graph()
    with eg.as_default():
        esess = tf.Session(config=tf.ConfigProto(allow_soft_placement=True))
        with tf.device('/gpu:0'):
            in_ph = [tf.placeholder(dtype=tf.float32, shape=[None, 1, 201], name='cleanph'),
                       tf.placeholder(dtype=tf.float32, shape=[None, Mix_Win, 201], name='mixedph'),
                       tf.placeholder(dtype=tf.float32, shape=[None, 1, 201], name='mixedphaseph'),
                       tf.placeholder(dtype=tf.float32, shape=[None, Noise_Win, 201], name='noisecontextph'),
                       tf.placeholder(dtype=tf.float32, shape=[None, Noise_Win, 201], name='cleancontextph'),
                       tf.placeholder(dtype=tf.int32, shape=[None], name='locationph'),
                       tf.placeholder(dtype=tf.string, shape=[None], name='cleanpathph'),
                       tf.placeholder(dtype=tf.string, shape=[None], name='noisepathph'),
                       tf.placeholder(dtype=tf.int32, shape=[None], name='snrph')]
            _, _, outputs = model(in_ph, False)
        esaver = tf.train.Saver(tf.global_variables())

    checkpoints_dir = './trained_model'
    esaver = tf.train.import_meta_graph(checkpoints_dir + '/81457_2-545000.meta')
    esaver.restore(esess, checkpoints_dir + '/=81457_2-545000')

    mixed_tensor = eg.get_tensor_by_name('mixedph:0')
    noise_tensor = eg.get_tensor_by_name('noisecontextph:0')
    clean_tensor = eg.get_tensor_by_name('cleancontextph:0')
    denoised_tensor = eg.get_tensor_by_name('add_72:0')

    # nn processing
    for i in range(batches):
        batch_mix_spectrum, batch_clean_spectrum, batch_noise_spectrum = [spectra[i*mb:(i+1)*mb] for spectra in [mix_spectra_, clean_spectra_, noise_spectra_]]

        batch_denoised_ = esess.run(denoised_tensor, feed_dict={mixed_tensor: batch_mix_spectrum,
                                                                noise_tensor: batch_noise_spectrum,
                                                                clean_tensor: batch_clean_spectrum})

        denoised.append(batch_denoised_)
        mix_center = batch_mix_spectrum[:, 17, :]
        mix_centers.append(mix_center)

    # reconstruction
    denoised = np.concatenate(denoised, axis=0)
    mix_centers = np.concatenate(mix_centers, axis=0)

    recover_samples_from_spectrum(denoised, mixedphs_[:, 0, :], save_to)
    mix_save_to = save_to[:-15] + 'mixed_demo.wav'
    recover_samples_from_spectrum(mix_centers, mixedphs_[:, 0, :], mix_save_to)


def apply_separator(mixedpath, cleanpath, noisepath, save_to):
    (mixedpath, cleanpath, noisepath), save_to_temp, fs, br = format_input_audio(mixedpath, cleanpath, noisepath, save_to)
    mixedlist = []
    mixedlist.append(mixedpath)
    cleanlist = []
    cleanlist.append(cleanpath)
    noiselist = []
    noiselist.append(noisepath)

    # data processing
    g = tf.Graph()
    with g.as_default():
        with tf.device('/cpu:0'):
            mixedlist = tf.constant(np.array(mixedlist))
            cleanlist = tf.constant(np.array(cleanlist))
            noiselist = tf.constant(np.array(noiselist))

            mixedpath_ph = mixedlist[0]
            cleanpath_ph = cleanlist[0]
            noisepath_ph = noiselist[0]

            clean_wav, noise_wav, mix_wav = tf.py_func(handle_signals,
                                                       [mixedpath_ph, cleanpath_ph, noisepath_ph],
                                                       [tf.float32, tf.float32, tf.float32])

            clean_wav, noise_wav, mix_wav = [tf.reshape(x, [-1]) for x in (clean_wav, noise_wav, mix_wav)]

            sess = tf.Session(config=tf.ConfigProto(allow_soft_placement=True))

            mix_stft, clean_stft, noise_stft = [tf.signal.stft(wav, 400, 160, fft_length=400) for wav in [mix_wav, clean_wav, noise_wav]]

            mix_spectrum, clean_spectrum, noise_spectrum = [tf.log(tf.abs(wav_stft) + 1e-5) for wav_stft in [mix_stft, clean_stft, noise_stft]]

            mix_phase = tf.angle(mix_stft)

            # crop data
            mix_spectra = strided_crop(mix_spectrum, Mix_Win, 1)

            # postive noise & negative noise
            clean_spectrum = clean_spectrum[:Noise_Win]
            clean_spectrum = tf.reshape(clean_spectrum, [Noise_Win, clean_spectrum.shape[1].value])
            clean_spectra = tf.tile(tf.expand_dims(clean_spectrum, 0), [tf.shape(mix_spectra)[0], 1, 1])

            noise_spectrum = noise_spectrum[:Noise_Win]
            noise_spectrum = tf.reshape(noise_spectrum, [Noise_Win, noise_spectrum.shape[1].value])
            noise_spectra = tf.tile(tf.expand_dims(noise_spectrum, 0), [tf.shape(mix_spectra)[0], 1, 1])

            mixedphs = strided_crop(mix_phase, 1, 1)

            # get data
            mix_spectra_, clean_spectra_, noise_spectra_ = sess.run([mix_spectra, clean_spectra, noise_spectra])
            mixedphs_ = sess.run(mixedphs, feed_dict={mixedpath_ph: mixedpath,
                                                      cleanpath_ph: cleanpath,
                                                      noisepath_ph: noisepath})

    mb = 100
    batches = int(math.ceil(len(mix_spectra_) / float(mb)))
    denoised = []
    mix_centers = []

    # denoiseing
    # graph
    eg = tf.Graph()
    with eg.as_default():
        esess = tf.Session(config=tf.ConfigProto(allow_soft_placement=True))
        with tf.device('/gpu:0'):
            in_ph = [tf.placeholder(dtype=tf.float32, shape=[None, 1, 201], name='cleanph'),
                       tf.placeholder(dtype=tf.float32, shape=[None, Mix_Win, 201], name='mixedph'),
                       tf.placeholder(dtype=tf.float32, shape=[None, 1, 201], name='mixedphaseph'),
                       tf.placeholder(dtype=tf.float32, shape=[None, Noise_Win, 201], name='noisecontextph'),
                       tf.placeholder(dtype=tf.float32, shape=[None, Noise_Win, 201], name='cleancontextph'),
                       tf.placeholder(dtype=tf.int32, shape=[None], name='locationph'),
                       tf.placeholder(dtype=tf.string, shape=[None], name='cleanpathph'),
                       tf.placeholder(dtype=tf.string, shape=[None], name='noisepathph'),
                       tf.placeholder(dtype=tf.int32, shape=[None], name='snrph')]
            _, _, outputs = model(in_ph, False)
        esaver = tf.train.Saver(tf.global_variables())

    checkpoints_dir = './trained_model'
    # checkpoints_dir = '/home/user/on_gpu/checkpoints/N_HANS___Source_Separation'
    esaver = tf.train.import_meta_graph(checkpoints_dir + '/81457_2-545000.meta')
    esaver.restore(esess, checkpoints_dir + '/81457_2-545000')

    mixed_tensor = eg.get_tensor_by_name('mixedph:0')
    noise_tensor = eg.get_tensor_by_name('noisecontextph:0')
    clean_tensor = eg.get_tensor_by_name('cleancontextph:0')
    denoised_tensor = eg.get_tensor_by_name('add_72:0')

    # nn processing
    for i in range(batches):
        batch_mix_spectrum, batch_clean_spectrum, batch_noise_spectrum = [spectra[i*mb:(i+1)*mb] for spectra in [mix_spectra_, clean_spectra_, noise_spectra_]]

        batch_denoised_ = esess.run(denoised_tensor, feed_dict={mixed_tensor: batch_mix_spectrum,
                                                                noise_tensor: batch_noise_spectrum,
                                                                clean_tensor: batch_clean_spectrum})

        denoised.append(batch_denoised_)
        mix_center = batch_mix_spectrum[:, 17, :]
        mix_centers.append(mix_center)

    # reconstruction
    denoised = np.concatenate(denoised, axis=0)
    mix_centers = np.concatenate(mix_centers, axis=0)

    recover_samples_from_spectrum(denoised, mixedphs_[:, 0, :], save_to_temp)

    # mix_save_to = save_to_temp[:-12] + 'mixed_processed.wav'
    # recover_samples_from_spectrum(mix_centers, mixedphs_[:, 0, :], mix_save_to)
    format_output_audio(save_to, fs, br)


def main():
    mixedpath = FLAGS.input
    noisepath = FLAGS.neg
    cleanpath = FLAGS.pos
    save_to = FLAGS.output

    if os.path.isdir(mixedpath) and os.path.isdir(cleanpath) and os.path.isdir(noisepath) and os.path.isdir(save_to):
        print('##########################################')
        print(' Applying N-HANS source separation system ')
        print('##########################################')
        for file in os.listdir(mixedpath):
            mixedwav = os.path.join(mixedpath, file)
            negwav = os.path.join(noisepath, file)
            poswav = os.path.join(cleanpath, file)
            savewav = os.path.join(save_to, file)
            print('Processing:')
            print('  Input : {}'.format(mixedwav))
            print('  +rec  : {}'.format(poswav))
            print('  -rec  : {}'.format(negwav))
            apply_separator(mixedwav, poswav, negwav, savewav)
            print('Processed result:')
            print('  Output: {}'.format(savewav))
            print('------------------------------------------')
            print()

    elif mixedpath.endswith('.wav') and cleanpath.endswith('.wav') and noisepath.endswith('.wav') and save_to.endswith('.wav'):
        print('##########################################')
        print(' Applying N-HANS source separation system ')
        print('##########################################')
        print('Processing:')
        print('  Input : {}'.format(mixedpath))
        print('  +rec  : {}'.format(cleanpath))
        print('  -rec  : {}'.format(noisepath))
        apply_separator(mixedpath, cleanpath, noisepath, save_to)
        print('Processed result:')
        print('  Output: {}'.format(save_to))
        print('------------------------------------------')
        print()

    else:
        print('Please give the .wav file path or the path to folders containing .wav files')

    # apply_separator(mixedpath, cleanpath, noisepath, save_to)


if __name__ == '__main__':
    # dir = '/nas/staff/data_work/Sure/example_wav/'
    mixedpath = FLAGS.input
    noisepath = FLAGS.neg
    cleanpath = FLAGS.pos
    save_to = FLAGS.output

    apply_separator(mixedpath, cleanpath, noisepath, save_to)









########################################################################################################################
#                                          N-HANS speech denoiser: apply                                               #
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
#   Description:      Apply N-HANS trained denoiser for audio.                          .                              #
#   Authors:          Shuo Liu, Gil Keren, Bjoern Schuller                                                             #
#   Affiliation:      Chair of Embedded Intelligence for Health Care and Wellbeing, University of Augsburg (UAU)       #
#   Version:          2.0                                                                                              #
#   Last Update:      June. 17, 2020                                                                                    #
#   Dependence Files: xxx                                                                                              #
#   Contact:          shuo.liu@informatik.uni-augburg.de                                                               #
########################################################################################################################
#                                                                                                                      #
#   Copyright (C) 2019, Shuo Liu, Gil Keren and Björn Schuller: University of Augsburg.                                #
#                                                                                                                      #
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public    #
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any       #
# later version.                                                                                                       #
#                                                                                                                      #
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied   #
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.#
#                                                                                                                      #
# You should have received a copy of the GNU General Public License along with this program. #  If not, see            #
# <http://www.gnu.org/licenses/>.                                                                                      #
########################################################################################################################

from __future__ import division, absolute_import, print_function
import numpy as np
import warnings

warnings.filterwarnings('ignore', category=FutureWarning)
import os
import math, functools
import sox
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
import tensorflow as tf
if int(tf.__version__.split('.')[0]) == 2:
    import tensorflow.compat.v1 as tf
    tf.compat.v1.disable_v2_behavior()
from tensorflow.python.util import deprecation
deprecation._PRINT_DEPRECATION_WARNINGS = False
tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
from scipy.io.wavfile import read as wavread
from scipy.io.wavfile import write as wavwrite
# from main import model
#         ||
#         \/
from N_HANS.N_HANS___Selective_Noise import main

model = main.model

FLAGS = tf.compat.v1.flags.FLAGS

tf.compat.v1.flags.DEFINE_string('trained_model_dir', './trained_model', '')
tf.compat.v1.flags.DEFINE_string('trained_model_name', '81448_0-1000000', '')

tf.compat.v1.flags.DEFINE_string('input', './audio_examples/exp2_noisy.wav', '')
tf.compat.v1.flags.DEFINE_string('neg', './audio_examples/exp2_noise.mp3', '')
tf.compat.v1.flags.DEFINE_string('pos', './audio_examples/Silent.wav', '')
tf.compat.v1.flags.DEFINE_string('output', './audio_examples/exp2_denoised.mp3', '')
tf.compat.v1.flags.DEFINE_float('compensate', 0, '')
tf.compat.v1.flags.DEFINE_boolean('ac', False, '')

tf.compat.v1.flags.DEFINE_string('temp_audio', './temp', '')

Noise_Win = 200
Mix_Win = 35

"""
Assistence Functions
"""


def format_input_audio(mixed, pos, neg, out):
    input_audio = [mixed, pos, neg]
    input_names = ['mixed', 'pos', 'neg']
    fs = int(sox.file_info.sample_rate(mixed))
    br = int(sox.file_info.bitrate(mixed))
    # check existence
    for idx, audio_file in enumerate([mixed, pos, neg]):
        if not os.path.exists(audio_file):
            print('+++++++++++++++++++++++++++++++++')
            print('Please check. {} path not exists.'.format(input_names[idx]))

    # converter
    soxer = sox.Transformer()
    soxer.convert(samplerate=16000, n_channels=1, bitdepth=16)

    for idx, audio_file in enumerate([mixed, pos, neg]):
        format_satisfied = audio_file.endswith('.wav')
        samplerate_satisfied = int(sox.file_info.sample_rate(audio_file)) == 16000
        satisfied = format_satisfied * samplerate_satisfied
        if format_satisfied:
            bitrate_satisfied = int(sox.file_info.bitrate(audio_file)) == 16
            satisfied = format_satisfied * samplerate_satisfied * bitrate_satisfied
        if not satisfied:
            if not os.path.exists('./temp'):
                os.mkdir('./temp')
            temp_file = './temp/' + input_names[idx] + '_temp.wav'
            soxer.build(audio_file, temp_file)
            input_audio[idx] = temp_file

    out_format = sox.file_info.file_extension(out)
    if out_format != 'wav':
        out = out.replace(out_format, 'wav')
    return input_audio, out, fs, br


def format_output_audio(out_file, fs, br):
    format = sox.file_info.file_extension(out_file)
    if format != 'wav':
        soxer = sox.Transformer()
        soxer.convert(samplerate=fs, bitdepth=br)
        soxer.build(out_file.replace(format, 'wav'), out_file)
        os.remove(out_file.replace(format, 'wav'))


def read_wav(in_path):
    rate, samples = wavread(in_path)
    assert rate == 16000
    assert samples.dtype == 'int16'
    if len(samples.shape) > 1:
        samples = samples.mean(axis=1)
    assert len(samples.shape) == 1
    return samples


def domixing(cleansamples, noisepossamples, noisenegsamples, snr_pos, snr_neg):
    # Repeat the noise if it is shorter than the speech, or shorter it if it's longer
    nse_pos = noisepossamples
    nse_neg = noisenegsamples
    while len(cleansamples) - len(nse_pos) > 0:  # Make noise longer
        diff = len(cleansamples) - len(nse_pos)
        nse_pos = np.concatenate([nse_pos, noisepossamples[:diff]], axis=0)

    while len(cleansamples) - len(nse_neg) > 0:  # Make noise longer
        diff = len(cleansamples) - len(nse_neg)
        nse_neg = np.concatenate([nse_neg, noisenegsamples[:diff]], axis=0)

    if len(cleansamples) - len(noisepossamples) < 0:  # Make noise shorter
        nse_pos = noisepossamples[:len(cleansamples)]

    if len(cleansamples) - len(noisenegsamples) < 0:  # Make noise shorter
        nse_neg = noisenegsamples[:len(cleansamples)]

    sig = cleansamples

    # Power of signal and noise
    psignal = sum(abs(sig) * abs(sig)) / sig.shape[0]
    pnoise_pos = sum(abs(nse_pos) * abs(nse_pos)) / nse_pos.shape[0]
    pnoise_neg = sum(abs(nse_neg) * abs(nse_neg)) / nse_neg.shape[0]

    # Compute scale factor
    if pnoise_pos == 0:
        K_pos = 1
    else:
        K_pos = (psignal / pnoise_pos) * pow(10, -snr_pos / 10.0)
        K_pos = np.sqrt(K_pos)

    if pnoise_neg == 0:
        K_neg = 1
    else:
        K_neg = (psignal / pnoise_neg) * pow(10, -snr_neg / 10.0)
        K_neg = np.sqrt(K_neg)

    # Mix
    noise_pos_scaled = K_pos * nse_pos  # Scale the noise
    noise_neg_scaled = K_neg * nse_neg  # Scale the noise
    mixed = sig + noise_pos_scaled + noise_neg_scaled  # Mix
    mixed = mixed / (max(abs(mixed)) + 0.000001)  # Normalize
    target = sig + noise_pos_scaled
    target = target / (max(abs(mixed)) + 0.000001)  # Normalize
    noise_pos_signal = noise_pos_scaled / (max(abs(mixed)) + 0.000001)
    noise_neg_signal = noise_neg_scaled / (max(abs(mixed)) + 0.000001)

    return mixed, target, K_pos, K_neg, noise_pos_signal, noise_neg_signal


def combine_signals(cleanpath, noisepospath, noisenegpath):
    try:
        # Read Wavs
        cleansamples = read_wav(cleanpath)
        noisepossamples = read_wav(noisepospath)
        noisenegsamples = read_wav(noisenegpath)

        # Normalize
        cleansamples = cleansamples / (max(abs(cleansamples)) + 0.000001)
        noisepossamples = noisepossamples / (max(abs(noisepossamples)) + 0.000001)
        noisenegsamples = noisenegsamples / (max(abs(noisenegsamples)) + 0.000001)
        cleansamples = cleansamples.astype(np.float32)
        noisepossamples = noisepossamples.astype(np.float32)
        noisenegsamples = noisenegsamples.astype(np.float32)

        # Cut the end to have an exact number of frames
        if (len(cleansamples) - 400) % 160 != 0:
            cleansamples = cleansamples[:-((len(cleansamples) - 400) % 160)]

        # Choose SNR
        SNRs = [-3, 0, 3, 5, 8]

        snr_pos = SNRs[1]
        snr_neg = SNRs[1]

        mixed, target, Kpos, Kneg, noisepossignal, noisenegsignal = domixing(cleansamples, noisepossamples,
                                                                             noisenegsamples, snr_pos, snr_neg)
        return noisepossignal, noisenegsignal, mixed, np.array(snr_pos, dtype=np.int32), np.array(snr_neg,
                                                                                                  dtype=np.int32)

    except:
        print('error in threads')
        print(cleanpath, noisepospath, noisenegpath)


def handle_signals(mixedpath, noisepospath, noisenegpath):
    try:
        # Read Wavs
        mixedsamples = read_wav(mixedpath)
        noisepossamples = read_wav(noisepospath)
        noisenegsamples = read_wav(noisenegpath)

        # Normalize
        mixedsamples = mixedsamples / (max(abs(mixedsamples)) + 0.000001)
        noisepossamples = noisepossamples / (max(abs(noisepossamples)) + 0.000001)
        noisenegsamples = noisenegsamples / (max(abs(noisenegsamples)) + 0.000001)
        mixedsamples = mixedsamples.astype(np.float32)
        noisepossamples = noisepossamples.astype(np.float32)
        noisenegsamples = noisenegsamples.astype(np.float32)

        desired_len = int(2.015*16000)
        while desired_len - len(noisepossamples) > 0:  # Make noise longer
            diff = desired_len - len(noisepossamples)
            noisepossamples = np.concatenate([noisepossamples, noisepossamples[:diff]], axis=0)

        while desired_len - len(noisenegsamples) > 0:  # Make noise longer
            diff = desired_len - len(noisenegsamples)
            noisenegsamples = np.concatenate([noisenegsamples, noisenegsamples[:diff]], axis=0)

        # Cut the end to have an exact number of frames
        if (len(mixedsamples) - 400) % 160 != 0:
            mixedsamples = mixedsamples[:-((len(mixedsamples) - 400) % 160)]

        return noisepossamples, noisenegsamples, mixedsamples

    except:
        print('error in threads')
        print(mixedpath, noisepospath, noisenegpath)


def pad_1D_for_windowing(tensor, length):
    len_before = ((length + 1) // 2) - 1
    len_after = length // 2
    return tf.pad(tensor, [[len_before, len_after], [0, 0]])


def strided_crop(tensor, length, stride):
    # we assume that we have a length dimension and a feature dimension
    assert len(tensor.shape) == 2
    n_features = int(tensor.shape[1])
    padded = pad_1D_for_windowing(tensor, length)
    windows = tf.extract_image_patches(tf.expand_dims(tf.expand_dims(padded, axis=0), axis=3),
                                       ksizes=[1, length, n_features, 1],
                                       strides=[1, stride, n_features, 1],
                                       rates=[1, 1, 1, 1],
                                       padding='VALID')
    return tf.reshape(windows, [-1, length, n_features])


def recover_samples_from_spectrum(logspectrum_stft, spectrum_phase, save_to=None):
    abs_spectrum = np.exp(logspectrum_stft)
    spectrum = abs_spectrum * (np.exp(1j * spectrum_phase))

    istft_graph = tf.Graph()
    with istft_graph.as_default():
        stft_ph = tf.placeholder(tf.complex64, shape=(None, 201))
        samples = tf.signal.inverse_stft(stft_ph, 400, 160, 400, window_fn=tf.signal.inverse_stft_window_fn(160, forward_window_fn=functools.partial(tf.signal.hann_window, periodic=True)))
        istft_sess = tf.Session(config=tf.ConfigProto(allow_soft_placement=True))
        samples_ = istft_sess.run(samples, feed_dict={stft_ph: spectrum})
        if save_to is not None:
            wavwrite(save_to, 16000, samples_)

    return samples_


"""
Apply N-HANS speech denoiser
"""


def apply_demo(speechpath, pospath, negpath, save_to):
    speechlist = []
    speechlist.append(speechpath)
    noiselist = []
    noiselist.append(pospath)
    noiselist.append(negpath)

    # data processing
    g = tf.Graph()
    with g.as_default():
        with tf.device('/cpu:0'):
            speechlist = tf.constant(np.array(speechlist))
            noiselist = tf.constant(np.array(noiselist))

            seeds = []
            seeds.append(speechlist)
            seeds.append(noiselist)

            speechpath_ph = speechlist[0]
            pospath_ph = noiselist[0]
            negpath_ph = noiselist[1]

            noise_pos_wav, noise_neg_wav, mix_wav, snr_pos, snr_neg = tf.py_func(combine_signals,
                                                                                 [speechpath_ph, pospath_ph,
                                                                                  negpath_ph],
                                                                                 [tf.float32,
                                                                                  tf.float32, tf.float32,
                                                                                  tf.int32, tf.int32])

            noise_pos_wav, noise_neg_wav, mix_wav = [tf.reshape(x, [-1]) for x in
                                                     (noise_pos_wav, noise_neg_wav, mix_wav)]

            sess = tf.Session(config=tf.ConfigProto(allow_soft_placement=True))

            mix_stft, pos_stft, neg_stft = [tf.signal.stft(wav, 400, 160, fft_length=400) for wav in
                                            [mix_wav, noise_pos_wav, noise_neg_wav]]

            mix_spectrum, pos_spectrum, neg_spectrum = [tf.log(tf.abs(wav_stft) + 1e-5) for wav_stft in
                                                        [mix_stft, pos_stft, neg_stft]]

            mix_phase = tf.angle(mix_stft)

            # crop data
            mix_spectra = strided_crop(mix_spectrum[Noise_Win:], Mix_Win, 1)

            # postive noise & negative noise
            pos_spectrum = pos_spectrum[:Noise_Win]
            pos_spectrum = tf.reshape(pos_spectrum, [Noise_Win, pos_spectrum.shape[1].value])
            pos_spectra = tf.tile(tf.expand_dims(pos_spectrum, 0), [tf.shape(mix_spectra)[0], 1, 1])

            neg_spectrum = neg_spectrum[:Noise_Win]
            neg_spectrum = tf.reshape(neg_spectrum, [Noise_Win, neg_spectrum.shape[1].value])
            neg_spectra = tf.tile(tf.expand_dims(neg_spectrum, 0), [tf.shape(mix_spectra)[0], 1, 1])

            mixedphs = strided_crop(mix_phase[Noise_Win:], 1, 1)

            # get data
            print('--------------------------------')
            mix_spectra_, pos_spectra_, neg_spectra_ = sess.run([mix_spectra, pos_spectra, neg_spectra])
            mixedphs_ = sess.run(mixedphs, feed_dict={speechpath_ph: speechpath,
                                                      pospath_ph: pospath,
                                                      negpath_ph: negpath})

    mb = 100
    batches = int(math.ceil(len(mix_spectra_) / float(mb)))
    denoised = []
    mix_centers = []

    # denoiseing
    # graph
    eg = tf.Graph()
    with eg.as_default():
        esess = tf.Session(config=tf.ConfigProto(allow_soft_placement=True))
        with tf.device('/gpu:0'):
            in_ph = [tf.placeholder(dtype=tf.float32, shape=[None, 1, 201], name='targetph'),
                     tf.placeholder(dtype=tf.float32, shape=[None, 35, 201], name='mixedph'),
                     tf.placeholder(dtype=tf.float32, shape=[None, 1, 201], name='mixedphaseph'),
                     tf.placeholder(dtype=tf.float32, shape=[None, 1, 201], name='targetphaseph'),
                     tf.placeholder(dtype=tf.float32, shape=[None, 35, 201], name='posph'),
                     tf.placeholder(dtype=tf.float32, shape=[None, 1, 201], name='posphaseph'),
                     tf.placeholder(dtype=tf.float32, shape=[None, 35, 201], name='negph'),
                     tf.placeholder(dtype=tf.float32, shape=[None, 1, 201], name='negphaseph'),
                     tf.placeholder(dtype=tf.float32, shape=[None, 200, 201], name='noiseposcontextph'),
                     tf.placeholder(dtype=tf.float32, shape=[None, 200, 201], name='noisenegcontextph'),
                     tf.placeholder(dtype=tf.int32, shape=[None], name='locationph'),
                     tf.placeholder(dtype=tf.string, shape=[None], name='targetpathph'),
                     tf.placeholder(dtype=tf.string, shape=[None], name='noisepospathph'),
                     tf.placeholder(dtype=tf.string, shape=[None], name='noisenegpathph'),
                     tf.placeholder(dtype=tf.int32, shape=[None], name='snrposph'),
                     tf.placeholder(dtype=tf.int32, shape=[None], name='snrnegph'),
                     ]

            _, _, outputs = model(in_ph, False)
        esaver = tf.train.Saver(tf.global_variables())

    checkpoints_dir = './trained_model'
    esaver = tf.train.import_meta_graph(checkpoints_dir + '/81448_0-1000000.meta')
    esaver.restore(esess, checkpoints_dir + '/81448_0-1000000')

    mixed_tensor = eg.get_tensor_by_name('mixedph:0')
    pos_noise_tensor = eg.get_tensor_by_name('noiseposcontextph:0')
    neg_noise_tensor = eg.get_tensor_by_name('noisenegcontextph:0')
    denoised_tensor = eg.get_tensor_by_name('add_72:0')

    # nn processing
    for i in range(batches):
        batch_mix_spectrum, batch_pos_spectrum, batch_neg_spectrum = [spectra[i * mb:(i + 1) * mb] for spectra in
                                                                      [mix_spectra_, pos_spectra_, neg_spectra_]]

        batch_denoised_ = esess.run(denoised_tensor, feed_dict={mixed_tensor: batch_mix_spectrum,
                                                                pos_noise_tensor: batch_pos_spectrum,
                                                                neg_noise_tensor: batch_neg_spectrum})

        denoised.append(batch_denoised_)
        mix_center = batch_mix_spectrum[:, 17, :]
        mix_centers.append(mix_center)

    # reconstruction
    denoised = np.concatenate(denoised, axis=0)
    mix_centers = np.concatenate(mix_centers, axis=0)

    recover_samples_from_spectrum(denoised, mixedphs_[:, 0, :], save_to)
    #mix_save_to = save_to[:-15] + 'mixed_demo.wav'
    #recover_samples_from_spectrum(mix_centers, mixedphs_[:, 0, :], mix_save_to)


def apply_snc(mixedpath, pospath, negpath, save_to):
    (mixedpath, pospath, negpath), save_to_temp, fs, br = format_input_audio(mixedpath, pospath, negpath, save_to)
    mixedlist = []
    mixedlist.append(mixedpath)
    posnoiselist = []
    negnoiselist = []
    posnoiselist.append(pospath)
    negnoiselist.append(negpath)

    # data processing
    g = tf.Graph()
    with g.as_default():
        with tf.device('/cpu:0'):
            mixedlist = tf.constant(np.array(mixedlist))
            posnoiselist = tf.constant(np.array(posnoiselist))
            negnoiselist = tf.constant(np.array(negnoiselist))

            mixedpath_ph = mixedlist[0]
            pospath_ph = posnoiselist[0]
            negpath_ph = negnoiselist[0]

            noise_pos_wav, noise_neg_wav, mix_wav = tf.py_func(handle_signals,
                                                               [mixedpath_ph, pospath_ph, negpath_ph],
                                                               [tf.float32, tf.float32, tf.float32])

            noise_pos_wav, noise_neg_wav, mix_wav = [tf.reshape(x, [-1]) for x in
                                                     (noise_pos_wav, noise_neg_wav, mix_wav)]

            sess = tf.Session(config=tf.ConfigProto(allow_soft_placement=True))

            mix_stft, pos_stft, neg_stft = [tf.signal.stft(wav, 400, 160, fft_length=400) for wav in
                                            [mix_wav, noise_pos_wav, noise_neg_wav]]

            mix_spectrum, pos_spectrum, neg_spectrum = [tf.log(tf.abs(wav_stft) + 1e-5) for wav_stft in
                                                        [mix_stft, pos_stft, neg_stft]]
            mix_phase = tf.angle(mix_stft)

            # crop data
            mix_spectra = strided_crop(mix_spectrum, Mix_Win, 1)

            # postive noise & negative noise
            pos_spectrum = pos_spectrum[:Noise_Win]
            pos_spectrum = tf.reshape(pos_spectrum, [Noise_Win, pos_spectrum.shape[1].value])
            pos_spectra = tf.tile(tf.expand_dims(pos_spectrum, 0), [tf.shape(mix_spectra)[0], 1, 1])

            neg_spectrum = neg_spectrum[:Noise_Win]
            neg_spectrum = tf.reshape(neg_spectrum, [Noise_Win, neg_spectrum.shape[1].value])
            neg_spectra = tf.tile(tf.expand_dims(neg_spectrum, 0), [tf.shape(mix_spectra)[0], 1, 1])

            mixedphs = strided_crop(mix_phase, 1, 1)

            # get data
            mix_spectra_, pos_spectra_, neg_spectra_ = sess.run([mix_spectra, pos_spectra, neg_spectra])
            mixedphs_ = sess.run(mixedphs, feed_dict={mixedpath_ph: mixedpath,
                                                      pospath_ph: pospath,
                                                      negpath_ph: negpath})

    mb = 100
    batches = int(math.ceil(len(mix_spectra_) / float(mb)))
    denoised = []
    mix_centers = []

    # denoiseing
    # graph
    eg = tf.Graph()
    with eg.as_default():
        esess = tf.Session(config=tf.ConfigProto(allow_soft_placement=True))
        with tf.device('/gpu:0'):
            in_ph = [tf.placeholder(dtype=tf.float32, shape=[None, 1, 201], name='targetph'),
                     tf.placeholder(dtype=tf.float32, shape=[None, 35, 201], name='mixedph'),
                     tf.placeholder(dtype=tf.float32, shape=[None, 1, 201], name='mixedphaseph'),
                     tf.placeholder(dtype=tf.float32, shape=[None, 1, 201], name='targetphaseph'),
                     tf.placeholder(dtype=tf.float32, shape=[None, 35, 201], name='posph'),
                     tf.placeholder(dtype=tf.float32, shape=[None, 1, 201], name='posphaseph'),
                     tf.placeholder(dtype=tf.float32, shape=[None, 35, 201], name='negph'),
                     tf.placeholder(dtype=tf.float32, shape=[None, 1, 201], name='negphaseph'),
                     tf.placeholder(dtype=tf.float32, shape=[None, 200, 201], name='noiseposcontextph'),
                     tf.placeholder(dtype=tf.float32, shape=[None, 200, 201], name='noisenegcontextph'),
                     tf.placeholder(dtype=tf.int32, shape=[None], name='locationph'),
                     tf.placeholder(dtype=tf.string, shape=[None], name='targetpathph'),
                     tf.placeholder(dtype=tf.string, shape=[None], name='noisepospathph'),
                     tf.placeholder(dtype=tf.string, shape=[None], name='noisenegpathph'),
                     tf.placeholder(dtype=tf.int32, shape=[None], name='snrposph'),
                     tf.placeholder(dtype=tf.int32, shape=[None], name='snrnegph'),
                     ]
            _, _, outputs = model(in_ph, False)
        esaver = tf.train.Saver(tf.global_variables())

    checkpoints_dir = './trained_model'
    esaver = tf.train.import_meta_graph(checkpoints_dir + '/81448_0-1000000.meta')
    esaver.restore(esess, checkpoints_dir + '/81448_0-1000000')

    mixed_tensor = eg.get_tensor_by_name('mixedph:0')
    pos_noise_tensor = eg.get_tensor_by_name('noiseposcontextph:0')
    neg_noise_tensor = eg.get_tensor_by_name('noisenegcontextph:0')
    denoised_tensor = eg.get_tensor_by_name('add_72:0')

    # nn processing
    for i in range(batches):
        batch_mix_spectrum, batch_pos_spectrum, batch_neg_spectrum = [spectra[i * mb:(i + 1) * mb] for spectra in
                                                                      [mix_spectra_, pos_spectra_, neg_spectra_]]

        batch_denoised_ = esess.run(denoised_tensor, feed_dict={mixed_tensor: batch_mix_spectrum,
                                                                pos_noise_tensor: batch_pos_spectrum,
                                                                neg_noise_tensor: batch_neg_spectrum})

        denoised.append(batch_denoised_)
        mix_center = batch_mix_spectrum[:, 17, :]
        mix_centers.append(mix_center)

    # reconstruction
    denoised = np.concatenate(denoised, axis=0)
    mix_centers = np.concatenate(mix_centers, axis=0)

    denoised_samples = recover_samples_from_spectrum(denoised, mixedphs_[:, 0, :], save_to_temp)
    # mix_save_to = save_to_temp[:-12] + 'mixed_processed.wav'
    mixed_samples = recover_samples_from_spectrum(mix_centers, mixedphs_[:, 0, :])
    # removed_save_to = save_to_temp[:-12] + 'removed.wav'
    removed_samples = mixed_samples - denoised_samples
    #wavwrite(removed_save_to, 16000, removed_samples)

    snr_est = 10 * np.log10(np.mean(np.square(denoised_samples)) / np.mean(np.square(removed_samples)))
    if not FLAGS.ac:
        compensation_factor = FLAGS.compensate
    else:
        compensation_factor = snr_est / 20
    # compensated_save_to = save_to_temp[:-12] + 'compensated.wav'
    # compensated_samples = denoised_samples + removed_samples * compensation_factor
    # wavwrite(compensated_save_to, 16000, compensated_samples)

    # snr = S/N   snr_est = S-e / N+e
    # e is distortion, need to be compensate. For a definite snr, Lower snr_est requires more compensation.
    format_output_audio(save_to, fs, br)
    return snr_est


def apply_denoiser(mixedpath, negpath, save_to):
    dir = './audio_examples/'
    pospath = dir + 'Silent.wav'
    snr = apply_snc(mixedpath, pospath, negpath, save_to)
    return snr


def main_denoiser():
    mixedpath = FLAGS.input
    negpath = FLAGS.neg
    save_to = FLAGS.output
    print(mixedpath, negpath, save_to)

    if os.path.isdir(mixedpath) and os.path.isdir(negpath) and os.path.isdir(save_to):
        print('##################################')
        print(' Applying N-HANS denoising system ')
        print('##################################')
        for file in os.listdir(mixedpath):
            mixedwav = os.path.join(mixedpath, file)
            negwav = os.path.join(negpath, file)
            savewav = os.path.join(save_to, file)
            print('Processing:')
            print('  Input : {}'.format(mixedwav))
            print('  -rec  : {}'.format(negwav))
            snr = apply_denoiser(mixedwav, negwav, savewav)
            print('Results:')
            print('  Output: {}'.format(savewav))
            print('  Estimated SNR: {:.2f}dB'.format(snr))
            print('----------------------------------')
            print()

    elif mixedpath.endswith('.wav') and negpath.endswith('.wav') and save_to.endswith('.wav'):
        print('##################################')
        print(' Applying N-HANS denoising system ')
        print('##################################')

        print('Processing:')
        print('  Input : {}'.format(mixedpath))
        print('  -rec  : {}'.format(negpath))
        snr = apply_denoiser(mixedpath, negpath, save_to)
        print('Processed result:')
        print('  Output: {}'.format(save_to))
        print('  Estimated SNR: {:.2f}dB'.format(snr))
        print('----------------------------------')
        print('')

    else:
        print('Please give the .wav file path or the path to folders containing .wav files')


def main_snc():
    mixedpath = FLAGS.input
    negpath = FLAGS.neg
    pospath = FLAGS.pos
    save_to = FLAGS.output

    if os.path.isdir(mixedpath) and os.path.isdir(pospath) and os.path.isdir(negpath) and os.path.isdir(save_to):
        print('####################################################')
        print(' Applying N-HANS selective noise suppression system ')
        print('####################################################')
        for file in os.listdir(mixedpath):
            mixedwav = os.path.join(mixedpath, file)
            negwav = os.path.join(negpath, file)
            poswav = os.path.join(pospath, file)
            savewav = os.path.join(save_to, file)
            print('Processing:')
            print('  Input : {}'.format(mixedwav))
            print('  +rec  : {}'.format(poswav))
            print('  -rec  : {}'.format(negwav))
            snr = apply_snc(mixedwav, poswav, negwav, savewav)
            print('Results:')
            print('  Output: {}'.format(savewav))
            print('  Estimated SNR: {:.2f}dB'.format(snr))
            print('----------------------------------------------------')
            print()

    elif mixedpath.endswith('.wav') and negpath.endswith('.wav') and save_to.endswith('.wav'):
        print('####################################################')
        print(' Applying N-HANS selective noise suppression system ')
        print('####################################################')
        print('Processing:')
        print('  Input : {}'.format(mixedpath))
        print('  +rec  : {}'.format(pospath))
        print('  -rec  : {}'.format(negpath))
        snr = apply_snc(mixedpath, pospath, negpath, save_to)
        print('Results:')
        print('  Output: {}'.format(save_to))
        print('  Estimated SNR: {:.2f}dB'.format(snr))
        print('----------------------------------------------------')
        print('')

    else:
        print('Please give the .wav file path or the path to folders containing .wav files')


def main():
    if FLAGS.pos == './audio_examples/Silent.wav':
        return main_denoiser()
    else:
        return main_snc()


if __name__ == '__main__':
    mixedpath = FLAGS.input
    negpath = FLAGS.neg
    save_to = FLAGS.output

    snr = apply_denoiser(mixedpath, negpath, save_to)



########################################################################################################################
#                                          N-HANS speech denoiser: blocks                                              #
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
#   Description:      Neural network blocks for constructing the N-HANS model.                                         #
#   Authors:          Shuo Liu, Gil Keren, Bjoern Schuller                                                             #
#   Affiliation:      Chair of Embedded Intelligence for Health Care and Wellbeing, University of Augsburg (UAU)       #
#   Version:          2.0                                                                                              #
#   Last Update:      May. 06, 2020                                                                                    #
#   Dependence Files: xxx                                                                                              #
#   Contact:          shuo.liu@informatik.uni-augburg.de                                                               #
########################################################################################################################
#                                                                                                                      #
#   Copyright (C) 2019, Shuo Liu, Gil Keren and Björn Schuller: University of Augsburg.                                #
#                                                                                                                      #
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public    #
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any       #
# later version.                                                                                                       #
#                                                                                                                      #
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied   #
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.#
#                                                                                                                      #
# You should have received a copy of the GNU General Public License along with this program. #  If not, see            #
# <http://www.gnu.org/licenses/>.                                                                                      #
########################################################################################################################

from __future__ import division, absolute_import, print_function
import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
import tensorflow as tf
from tensorflow.python.util import deprecation

deprecation._PRINT_DEPRECATION_WARNINGS = False
tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)

FLAGS = tf.compat.v1.flags.FLAGS
#rnn_cell = tf.nn.rnn_cell
#static_rnn = tf.nn.static_rnn


def dense(tensor, out_dim, w_std, b_init, with_bias, scope_name, bias_initializer=None):
    in_dim = tensor.get_shape()[1].value

    with tf.compat.v1.variable_scope(scope_name):
        w = tf.compat.v1.get_variable('w', [in_dim, out_dim], initializer=tf.compat.v1.truncated_normal_initializer(stddev=w_std))
        out = tf.matmul(tensor, w)
        if with_bias:
            if bias_initializer is None:
                b = tf.compat.v1.get_variable('b', [1, out_dim], initializer=tf.constant_initializer(b_init))
            else:
                b = tf.compat.v1.get_variable('b', [1, out_dim], initializer=bias_initializer)
            out += b
        return out


def conv2d(tensor, kernel_shape, strides, out_channels, w_std, b_init, with_bias, padding, scope_name):
    in_channels = tensor.get_shape()[3].value

    with tf.compat.v1.variable_scope(scope_name):
        kernel = tf.compat.v1.get_variable('w', kernel_shape + [in_channels, out_channels],
                                           initializer=tf.compat.v1.truncated_normal_initializer(stddev=w_std))
        out = tf.nn.conv2d(tensor, kernel, strides, padding=padding)
        if with_bias:
            b = tf.compat.v1.get_variable('b', [1, 1, 1, out_channels], initializer=tf.constant_initializer(b_init))
            out += b
        return out


# Input of shape [mb, time, features]
# def lstm(tensor, nlayers, nunits, w_std):
#     with tf.compat.v1.variable_scope('rec_cell', initializer=tf.random_normal_initializer(0.0, w_std)):
#         if nlayers == 1:
#             cell = rnn_cell.LSTMCell(num_units=nunits)
#         else:
#             cell = rnn_cell.MultiRNNCell([rnn_cell.LSTMCell(num_units=nunits) for _ in range(0, nlayers)])
#
#     inputs_as_list = tf.unstack(tf.transpose(tensor, [1, 0, 2]))
#     outputs, states = static_rnn(cell, inputs_as_list, dtype=tf.float32)
#     return outputs, states


def flatten(tensor):
    sh = tensor.get_shape()
    other_dimensions = 1
    for i in range(1, len(sh)):
        other_dimensions *= sh[i].value
    return tf.reshape(tensor, [-1, other_dimensions])


def batch_norm(is_train, tensor, scope_name):
    with tf.compat.v1.variable_scope(scope_name):
        BATCHNORM_MOVING_AVERAGE_DECAY = FLAGS.bn_decay
        # Get the shape of mean, variance, beta, gamma
        mask_shape = [1] * len(tensor.get_shape())
        mask_shape[-1] = tensor.get_shape()[-1].value

        # Create trainable variables to hold beta and gamma
        beta = tf.compat.v1.get_variable('beta', mask_shape, initializer=tf.constant_initializer(0.0))
        gamma = tf.compat.v1.get_variable('gamma', mask_shape, initializer=tf.constant_initializer(1.0))

        # Create non-trainable variables for the population mean and variance, and add them to the
        pop_mean = tf.compat.v1.get_variable('pop_mean', mask_shape, initializer=tf.constant_initializer(0.0),
                                             trainable=False)
        pop_variance = tf.compat.v1.get_variable('pop_variance', mask_shape, initializer=tf.constant_initializer(1.0),
                                                 trainable=False)

        if is_train:
            # Calculate the moments based on the individual batch.
            n_dims = len(tensor.get_shape())
            mean, variance = tf.nn.moments(x=tensor, axes=[i for i in range(0, n_dims - 1)], keep_dims=True)

            # Update the population mean and variance
            pop_mean_update = pop_mean.assign(
                (BATCHNORM_MOVING_AVERAGE_DECAY * pop_mean + (1 - BATCHNORM_MOVING_AVERAGE_DECAY) * mean))
            pop_variance_update = pop_variance.assign(
                (BATCHNORM_MOVING_AVERAGE_DECAY * pop_variance + (1 - BATCHNORM_MOVING_AVERAGE_DECAY) * variance))

            # Add the update ops to a collection, in order to perform them when training
            with tf.control_dependencies([pop_mean_update, pop_variance_update]):
                return tf.nn.batch_normalization(tensor, mean, variance, beta, gamma, variance_epsilon=0.001)

        else:
            # Just use the moving_mean and moving_variance.
            mean = pop_mean
            variance = pop_variance
            return tf.nn.batch_normalization(tensor, mean, variance, beta, gamma, variance_epsilon=0.001)


def reverse_gradient(tensor, gradient_factor):
    return -gradient_factor * tensor + tf.stop_gradient((gradient_factor + 1) * tensor)



########################################################################################################################
#                                          N-HANS speech separator: reader                                             #
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
#   Description:      Data reader pipeline.                                                                            #
#   Authors:          Shuo Liu, Gil Keren, Bjoern Schuller                                                             #
#   Affiliation:      Chair of Embedded Intelligence for Health Care and Wellbeing, University of Augsburg (UAU)       #
#   Version:          2.0                                                                                              #
#   Last Update:      June. 17, 2020                                                                                   #
#   Dependence Files: xxx                                                                                              #
#   Contact:          shuo.liu@informatik.uni-augburg.de                                                               #
########################################################################################################################
#                                                                                                                      #
#   Copyright (C) 2019, Shuo Liu, Gil Keren and Björn Schuller: University of Augsburg.                                #
#                                                                                                                      #
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public    #
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any       #
# later version.                                                                                                       #
#                                                                                                                      #
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied   #
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.#
#                                                                                                                      #
# You should have received a copy of the GNU General Public License along with this program. #  If not, see            #
# <http://www.gnu.org/licenses/>.                                                                                      #
#                                                                                                                      #
########################################################################################################################

import numpy as np
import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
import tensorflow as tf
from tensorflow.python.util import deprecation
deprecation._PRINT_DEPRECATION_WARNINGS = False
tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
import random
import time
import hashlib
import pickle as p
from scipy.io.wavfile import read as wavread


"""
context_frames  : number of context frames          (reference signal)
window_frames   : number of frames of noisy signal  (input signal) 
random_slices   : number of slices (each slice contains one input signal and reference signals) 
eval_seeds      : 'valid' or 'test'. In training, evaluation is applied for 'valid' dataset, and in test, for 'test' dataset
wav_dump_folder : the folder to save denoised signals
"""
FLAGS = tf.compat.v1.flags.FLAGS
tf.compat.v1.flags.DEFINE_integer('context_frames', 200, '')
tf.compat.v1.flags.DEFINE_integer('window_frames', 35, '')
tf.compat.v1.flags.DEFINE_integer('random_slices', 50, '')
tf.compat.v1.flags.DEFINE_string('eval_seeds', 'valid', '')
tf.compat.v1.flags.DEFINE_string('wav_dump_folder', './wav_dump/', '')
tf.compat.v1.flags.DEFINE_string('speech_wav_dir', './speech_wav_dir/', '')


"""
Create Pickle List for Training, Evaluation and Test
"""


def create_seeds():
    """
    Create 3 pickle files: train.pkl, valid.pkl and test.pkl for speech audio files
    :param speech_wav_dir: the directory contains three subfolders, train, valid and test, each includes audio .wav files
    """
    speech_wav_dir = FLAGS.speech_wav_dir
    train_folder_clean = speech_wav_dir + 'train'
    valid_folder_clean = speech_wav_dir + 'valid'
    test_folder_clean = speech_wav_dir + 'test'
    trainseeds = []
    validseeds = []
    testseeds = []
    wavfolders = [train_folder_clean, valid_folder_clean, test_folder_clean]
    seedsfolders = [trainseeds, validseeds, testseeds]

    for ii in range(3):
        wavfolder = wavfolders[ii]
        seedfolder = seedsfolders[ii]
        for root, directories, filenames in os.walk(wavfolder):
            for filename in filenames:
                if filename.endswith('.wav'):
                    seedfolder.append(os.path.join(root, filename))

    train_pkl = speech_wav_dir + 'train.pkl'
    valid_pkl = speech_wav_dir + 'valid.pkl'
    test_pkl = speech_wav_dir + 'test.pkl'
    p.dump(trainseeds, open(train_pkl, "wb"))
    p.dump(validseeds, open(valid_pkl, "wb"))
    p.dump(testseeds, open(test_pkl, "wb"))


"""
Additional Functions for Generating Audio Samples
combine_signals: generating training, valid and test .wav files
"""


def read_wav(in_path):
    rate, samples = wavread(in_path)
    assert rate == 16000
    assert samples.dtype == 'int16'
    if len(samples.shape) > 1:
        samples = samples.mean(axis=1)
    assert len(samples.shape) == 1
    return samples


def domixing(cleansamples, noisesamples, snr):
    # Repeat the noise if it is shorter than the speech, or shorter it if it's longer
    nse = noisesamples
    while len(cleansamples) - len(nse) > 0:  # Make noise longer
        diff = len(cleansamples) - len(nse)
        nse = np.concatenate([nse, noisesamples[:diff]], axis=0)
    if len(cleansamples) - len(noisesamples) < 0:  # Make noise shorter
        nse = noisesamples[:len(cleansamples)]
    sig = cleansamples

    # Power of signal and noise
    psignal = sum(abs(sig) * abs(sig)) / sig.shape[0]
    pnoise = sum(abs(nse) * abs(nse)) / nse.shape[0]

    # Compute scale factor
    if pnoise == 0:
        K = 1
    else:
        K = (psignal / pnoise) * pow(10, -snr / 10.0)
    K = np.sqrt(K)

    # Mix
    noise_scaled = K * nse  # Scale the noise
    mixed = sig + noise_scaled  # Mix
    mixed = mixed / (max(abs(mixed)) + 0.000001)  # Normalize
    return mixed, K


def combine_signals(istrain, cleanpath, noisepath):
    try:
        # Read Wavs
        cleansamples = read_wav(cleanpath)
        noisesamples = read_wav(noisepath)

        # Normalize
        cleansamples = cleansamples / (max(abs(cleansamples)) + 0.000001)
        noisesamples = noisesamples / (max(abs(noisesamples)) + 0.000001)
        cleansamples = cleansamples.astype(np.float32)
        noisesamples = noisesamples.astype(np.float32)

        # Cut the end to have an exact number of frames
        cleansamples = cleansamples[:-((len(cleansamples) - 400) % 160)]

        # Choose SNR
        SNRs = [-5, -3, -1, 0, 1, 3, 5]

        if istrain:
            snr = SNRs[random.randint(0, len(SNRs) - 1)]
            mixed, K = domixing(cleansamples, noisesamples, snr)
            return cleansamples, noisesamples * K, mixed, np.array(snr, dtype=np.int32)

        else:
            # The noise is the beginning of the file, and mix the rest
            snrid = int(hashlib.md5(cleanpath).hexdigest()[:8], 16) % len(SNRs)
            snr = SNRs[snrid]
            mixed, K = domixing(cleansamples, noisesamples, snr)
            return cleansamples, noisesamples * K, mixed, np.array(snr, dtype=np.int32)
    except:
        print('error in threads')
        print(cleanpath, noisepath)


"""
Global FLAGS and Training FLAGS
"""


class read_seeds:
    """
    name              : 'train', 'valid' or 'test'. To generate data reader.
    queuesize         : the size of data queue
    min_after_dequeue : minimum seeds left in queue after dequeue operation
    nthreads          : number of threads processing dequeue
    """

    def __init__(self, name, queuesize, min_after_dequeue, nthreads):
        self.Fs = 16000
        self.frame_length = int(self.Fs * 0.025)
        self.frame_step = int(self.Fs * 0.010)
        self.queuesize = queuesize
        self.min_after_dequeue = min_after_dequeue
        self.window_frames = FLAGS.window_frames
        self.random_slices = FLAGS.random_slices
        self.context_frames = FLAGS.context_frames
        self.eval_stride = 1
        self.istrain = name == 'train'
        self.nthreads = nthreads
        self.name = name

        if self.istrain != True:
            name = FLAGS.eval_seeds

        speech_wav_dir = FLAGS.speech_wav_dir
        if name == 'train':
            self.seedspath = speech_wav_dir + 'train.pkl'
        elif name == 'valid':
            self.seedspath = speech_wav_dir + 'valid.pkl'
        elif name == 'test':
            self.seedspath = speech_wav_dir + 'test.pkl'

    def preparations(self):
        """
        Preparation of the data queue.
        """
        self.seeds = []
        with open(self.seedspath, 'rb') as f:
            seedlist = p.load(f)
            seedlist = tf.constant(np.array(seedlist))
            self.seeds.append(seedlist)

        # Construct seeds queues
        self.seedsqs = []
        for i, seed in enumerate(self.seeds):
            num_epochs = None if self.istrain else 1
            seedsq = tf.train.input_producer(seed, capacity=50000, shuffle=self.istrain, num_epochs=num_epochs)
            self.epochcounter = None if self.istrain else tf.compat.v1.local_variables()[-1]
            self.seedsqs.append(seedsq)

        # Get the examples from seeds. It is a list of tensors, with first dimension being the example.
        examples = self.get_examples(self.istrain, self.seedsqs)

        # Create examples queue and enqueue the examples
        dtypes = [x.dtype for x in examples]
        shapes = [x.shape.as_list()[1:] for x in examples]
        if self.istrain:
            exq = tf.queue.RandomShuffleQueue(capacity=self.queuesize, min_after_dequeue=self.min_after_dequeue,
                                              dtypes=dtypes, shapes=shapes)
        else:
            exq = tf.queue.FIFOQueue(capacity=self.queuesize, dtypes=dtypes, shapes=shapes)
        enqop = exq.enqueue_many(examples)

        # The dequeue operation and the mb placeholder
        mb = tf.compat.v1.placeholder(dtype=tf.int32, shape=[], name=self.name + '_mb')
        if self.istrain:
            batch = exq.dequeue_many(mb)
        else:
            batch = exq.dequeue_up_to(mb)

        # Create queue runner
        qr = tf.train.QueueRunner(exq, [enqop] * self.nthreads)
        tf.train.add_queue_runner(qr, tf.compat.v1.GraphKeys.QUEUE_RUNNERS)

        # Set some members
        self.exq = exq
        self.batch = batch
        self.mb = mb

    def get_examples(self, istrain, seedsqs):
        """
        get model inputs from audio samples
        :param istrain: boolean value. If True, the examples are randomly trauncated from audio samples; If False, the examples are trauncated frame by frame from audio samples
        :param seedsqs: seed list of speech files
        :return:
        """
        target_seed = seedsqs[0].dequeue()
        noise_seed = seedsqs[0].dequeue()
        target_wav, noise_wav, mix_wav, snr = tf.py_func(combine_signals, [istrain, target_seed, noise_seed],
                                                         [tf.float32, tf.float32, tf.float32, tf.int32])

        # flatten
        target_wav, noise_wav, mix_wav = [tf.reshape(x, [-1]) for x in (target_wav, noise_wav, mix_wav)]

        # stft
        target_fft = tf.signal.stft(target_wav, self.frame_length, self.frame_step,
                                            fft_length=self.frame_length)
        noise_fft = tf.signal.stft(noise_wav, self.frame_length, self.frame_step, fft_length=self.frame_length)
        mix_fft = tf.signal.stft(mix_wav, self.frame_length, self.frame_step, fft_length=self.frame_length)

        # Magnitude
        target_mag, noise_mag, mix_mag = [tf.abs(x) for x in [target_fft, noise_fft, mix_fft]]
        # Phase
        mix_phase = tf.math.angle(mix_fft)
        # log
        target_log, noise_log, mix_log = [tf.math.log(x + 1e-5) for x in [target_mag, noise_mag, mix_mag]]

        if istrain:
            # padding
            target_log, noise_log, mix_log, mix_phase = [self.pad_1D_for_windowing(x, self.window_frames) for x in
                                                         [target_log, noise_log, mix_log, mix_phase]]

            # Get n random window-label pairs
            targetslices = []
            mixedslices = []
            mixedphslices = []
            noise_context_slices = []
            clean_context_slices = []
            for _ in range(self.random_slices):
                # Crop a window
                cleancropped, mixedcropped, mixedphcropped, noisecontext, cleancontext = \
                    self.synchronized_1D_crop(target_log, noise_log, mix_log, mix_phase, self.window_frames,
                                              self.context_frames)

                # Add to list
                for (lst, tnsr) in zip(
                        [targetslices, mixedslices, mixedphslices, noise_context_slices, clean_context_slices],
                        [cleancropped, mixedcropped, mixedphcropped, noisecontext, cleancontext]):
                    lst.append(tnsr)

            # Stack and return
            return tf.stack(targetslices, axis=0), tf.stack(mixedslices, axis=0), tf.stack(mixedphslices, axis=0), \
                   tf.stack(noise_context_slices, axis=0), tf.stack(clean_context_slices, axis=0), \
                   tf.constant([0] * self.random_slices, dtype=tf.int32), tf.constant([''] * self.random_slices,
                                                                                      dtype=tf.string), \
                   tf.constant([''] * self.random_slices, dtype=tf.string), tf.constant([0] * self.random_slices,
                                                                                        dtype=tf.int32)
        else:
            targetslices = self.strided_crop(target_log[self.context_frames:], 1, self.eval_stride)
            mixedslices = self.strided_crop(mix_log[self.context_frames:], self.window_frames, self.eval_stride)
            mixedphslices = self.strided_crop(mix_phase[self.context_frames:], 1, self.eval_stride)
            noisecontext = noise_log[:self.context_frames]
            noisecontext = tf.reshape(noisecontext, [self.context_frames, noisecontext.shape[1].value])
            noise_context_slices = tf.tile(tf.expand_dims(noisecontext, 0), [tf.shape(targetslices)[0], 1, 1])
            cleancontext = target_log[:self.context_frames]
            cleancontext = tf.reshape(cleancontext, [self.context_frames, cleancontext.shape[1].value])
            clean_context_slices = tf.tile(tf.expand_dims(cleancontext, 0), [tf.shape(targetslices)[0], 1, 1])
            locations = tf.range(0, tf.shape(targetslices)[0], dtype=tf.int32)
            target_seeds = tf.tile([target_seed], [tf.shape(targetslices)[0]])
            noise_seeds = tf.tile([noise_seed], [tf.shape(targetslices)[0]])
            snrs = tf.tile([snr], [tf.shape(targetslices)[0]])
            return targetslices, mixedslices, mixedphslices, noise_context_slices, clean_context_slices, locations, target_seeds, noise_seeds, snrs

    def pad_1D_for_windowing(self, tensor, length):
        """
        Padding the input spectrogramm
        :param tensor: shape of [Times x Frequencies]
        :param length: equals to window frames
        :return: padded spectrogram
        """
        len_before = ((length + 1) // 2) - 1
        len_after = length // 2
        return tf.pad(tensor, [[len_before, len_after], [0, 0]])

    def synchronized_1D_crop(self, clean, noise, mixed, mixedph, winlength, contextlength):
        winmaxval = tf.shape(mixed)[0] - winlength
        winstart = tf.random.uniform([], dtype=tf.int32, minval=0, maxval=winmaxval + 1)

        # Crop window
        mixedcropped = mixed[winstart:winstart + winlength]
        cleancropped = clean[winstart:winstart + winlength]
        mixedphcropped = mixedph[winstart:winstart + winlength]
        mixedcropped = tf.reshape(mixedcropped, [winlength, mixedcropped.shape[1].value])
        cleancropped = tf.reshape(cleancropped, [winlength, cleancropped.shape[1].value])
        mixedphcropped = tf.reshape(mixedphcropped, [winlength, mixedphcropped.shape[1].value])

        # Get the center of the clean and mixed phase
        cleancropped = cleancropped[self.window_frames // 2]
        cleancropped = tf.expand_dims(cleancropped, 0)
        mixedphcropped = mixedphcropped[self.window_frames // 2]
        mixedphcropped = tf.expand_dims(mixedphcropped, 0)

        # The rest of the noise
        cleanrest = tf.concat([clean[:winstart], clean[winstart + winlength:]], axis=0)
        noiserest = tf.concat([noise[:winstart], noise[winstart + winlength:]], axis=0)

        # Get the noise part
        restmaxval = tf.shape(noiserest)[0] - contextlength
        reststart = tf.random.uniform([], dtype=tf.int32, minval=0, maxval=restmaxval + 1)
        noisecontext = noiserest[reststart:reststart + contextlength]
        noisecontext = tf.reshape(noisecontext, [contextlength, noisecontext.shape[1].value])

        restmaxval = tf.shape(cleanrest)[0] - contextlength
        reststart = tf.random.uniform([], dtype=tf.int32, minval=0, maxval=restmaxval + 1)
        cleancontext = cleanrest[reststart:reststart + contextlength]
        cleancontext = tf.reshape(cleancontext, [contextlength, cleancontext.shape[1].value])

        return cleancropped, mixedcropped, mixedphcropped, noisecontext, cleancontext

    def strided_crop(self, tensor, length, stride):
        # we assume that we have a length dimension and a feature dimension
        assert len(tensor.shape) == 2
        n_features = int(tensor.shape[1])
        padded = self.pad_1D_for_windowing(tensor, length)
        windows = tf.extract_image_patches(tf.expand_dims(tf.expand_dims(padded, axis=0), axis=3),
                                           ksizes=[1, length, n_features, 1],
                                           strides=[1, stride, n_features, 1],
                                           rates=[1, 1, 1, 1],
                                           padding='VALID')
        return tf.reshape(windows, [-1, length, n_features])

    def get_inputs(self):
        if not hasattr(self, 'ph'):
            self.ph = [tf.compat.v1.placeholder(dtype=tf.float32, shape=[None, 1, 201], name='cleanph'),
                       tf.compat.v1.placeholder(dtype=tf.float32, shape=[None, self.window_frames, 201],
                                                name='mixedph'),
                       tf.compat.v1.placeholder(dtype=tf.float32, shape=[None, 1, 201], name='mixedphaseph'),
                       tf.compat.v1.placeholder(dtype=tf.float32, shape=[None, self.context_frames, 201],
                                                name='noisecontextph'),
                       tf.compat.v1.placeholder(dtype=tf.float32, shape=[None, self.context_frames, 201],
                                                name='cleancontextph'),
                       tf.compat.v1.placeholder(dtype=tf.int32, shape=[None], name='locationph'),
                       tf.compat.v1.placeholder(dtype=tf.string, shape=[None], name='cleanpathph'),
                       tf.compat.v1.placeholder(dtype=tf.string, shape=[None], name='noisepathph'),
                       tf.compat.v1.placeholder(dtype=tf.int32, shape=[None], name='snrph')]
        return self.ph


if __name__ == '__main__':
    g = tf.Graph()
    with g.as_default():
        reader = read_seeds(name='train', queuesize=200, min_after_dequeue=100, nthreads=16)
        with tf.device('/cpu:0'):
            reader.preparations()

        print('starting queue runners')
        sess = tf.compat.v1.Session(graph=g, config=tf.ConfigProto(allow_soft_placement=True))
        sess.run(tf.compat.v1.local_variables_initializer())
        coord = tf.train.Coordinator()
        threads = tf.train.start_queue_runners(sess=sess, coord=coord)
        queue_stats = tf.stack([x.queue.size() for x in tf.get_collection(tf.compat.v1.GraphKeys.QUEUE_RUNNERS)])
        feed_dict = {reader.mb: 64}

        with tf.device('/gpu:0'):
            clean, mixed, mixedph, noisecontext, cleancontext, location, cleanpath, noisepath, snr = reader.get_inputs()

        for i in range(1000):
            print(i)
            t1 = time.time()
            clean_, mixed_, mixedph_, noisecontext_, cleancontext_, location_, cleanpath_, noisepath_, snr_ = sess.run(
                reader.batch, feed_dict=feed_dict)
            t2 = time.time()
            print(sess.run(queue_stats), t2 - t1)



########################################################################################################################
#                                          N-HANS speech separator                                                     #
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
#   Description:      N-HANS contains two systems, speech denoiser and speech separator. This is the first part.       #
#   Authors:          Shuo Liu, Gil Keren, Bjoern Schuller                                                             #
#   Affiliation:      Chair of Embedded Intelligence for Health Care and Wellbeing, University of Augsburg (UAU)  #
#   Version:          2.0                                                                                              #
#   Last Update:      May. 06, 2020                                                                                    #
#   Dependence Files: reader.py  blocks.py                                                                             #
#   Contact:          shuo.liu@informatik.uni-augburg.de                                                               #
########################################################################################################################
#                                                                                                                      #
#   Copyright (C) 2019, Shuo Liu, Gil Keren and Björn Schuller: University of Augsburg.                                #
#                                                                                                                      #
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public    #
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any       #
# later version.                                                                                                       #
#                                                                                                                      #
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied   #
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.#
#                                                                                                                      #
# You should have received a copy of the GNU General Public License along with this program. #  If not, see            #
# <http://www.gnu.org/licenses/>.                                                                                      #
#                                                                                                                      #
########################################################################################################################

# Import from standard libraries
from __future__ import division, absolute_import, print_function
import os, time, numpy as np, scipy.io.wavfile, functools
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
import warnings
warnings.filterwarnings('ignore', category=FutureWarning)
import tensorflow as tf
if int(tf.__version__.split('.')[0]) == 2:
    import tensorflow.compat.v1 as tf
    tf.compat.v1.disable_v2_behavior()
from tensorflow.python.util import deprecation
deprecation._PRINT_DEPRECATION_WARNINGS = False
tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
from datetime import datetime
from os.path import join
# from blocks import dense, conv2d, batch_norm, flatten
# from reader import read_seeds
#               ||
#               \/
from N_HANS.N_HANS___Source_Separation import blocks, reader
dense, conv2d, batch_norm, flatten = blocks.dense, blocks.conv2d, blocks.batch_norm, blocks.flatten
read_seeds = reader.read_seeds


# General flags
"""
During Training     : eval_before_train => False, eval_after_train => True
During Test         : eval_before_train => True, eval_after_train => False
restore_path        : the path to restore a trained model
checkpoints         : checkpoints directory
summaries           : summaries directory
dump_results        : directory of intermediate output of models
eval_every          : evaluation every "eval_every" batches
train_monitor_every : monitor training process every "train_monitor_every" batches
"""
FLAGS = tf.compat.v1.flags.FLAGS
tf.compat.v1.flags.DEFINE_boolean('eval_before_training', False, '')
tf.compat.v1.flags.DEFINE_boolean('eval_after_training', True, '')
tf.compat.v1.flags.DEFINE_integer('checkpoints_to_keep', 1000000, '')
tf.compat.v1.flags.DEFINE_string('restore_path', '', '')
tf.compat.v1.flags.DEFINE_string('model_name', 'nhans', '')
tf.compat.v1.flags.DEFINE_string('checkpoint_dir', './checkpoints', '')
tf.compat.v1.flags.DEFINE_string('summaries_dir', './summaries', '')
tf.compat.v1.flags.DEFINE_string('dump_results', './dump', '')
tf.compat.v1.flags.DEFINE_integer('eval_every', 3, '')  # 5000
tf.compat.v1.flags.DEFINE_integer('train_monitor_every', 3, '')  # 1000

# Training and model flags
"""
batches       : maximum batches for training
alg           : optimiser
lr            : learning rate
mom           : momentum for optimiser
w_std, b_init : parameters initialisation
bn_decay      : batch normalisation decay
train_mb      : mini-batch size for training
eval_mb       : mini-batch size for valid or test
"""
tf.compat.v1.flags.DEFINE_integer('batches', 3000000, '')
tf.compat.v1.flags.DEFINE_string('alg', 'sgd', '')
tf.compat.v1.flags.DEFINE_float('lr', 0.001, '')
tf.compat.v1.flags.DEFINE_float('mom', 0.0, '')
tf.compat.v1.flags.DEFINE_float('w_std', 0.01, '')
tf.compat.v1.flags.DEFINE_float('b_init', 0.0, '')
tf.compat.v1.flags.DEFINE_float('bn_decay', 0.95, '')
tf.compat.v1.flags.DEFINE_integer('train_mb', 64, '')  # 64
tf.compat.v1.flags.DEFINE_integer('eval_mb', 100, '')  # 100

"""
Preparation of Data Reader
"""

global treader, tsess, tph, modelname, isess, tin


def get_train_reader():
    return read_seeds('train', 40000, 6666, 16)  # 10, 1, 2


def get_eval_readers():
    return [read_seeds('valid', 5000, 0, 1)]  # 10, 0, 1


"""
N-HANS Separator Model
"""


def model(inputs, istrain):
    clean, mixed, mixedph, noisecontext, cleancontext, location, cleanpath, noisepath, snr = inputs
    nfeat = clean.shape[2].value

    def noise_resnet_block(inputs, kernel_size, stride, n_fmaps, scope_name):
        """
        Embedding network
        :param inputs : context input
        :return out   : embedding vector represents the context information
        """
        # The transformation path
        path1 = conv2d(inputs, kernel_size, [1] + stride + [1], n_fmaps, FLAGS.w_std, FLAGS.b_init, False, 'SAME',
                       scope_name + '_conv1')
        path1 = batch_norm(istrain, path1, scope_name + '_conv1')
        path1 = tf.nn.relu(path1)
        path1 = conv2d(path1, kernel_size, [1, 1, 1, 1], n_fmaps, FLAGS.w_std, FLAGS.b_init, True, 'SAME',
                       scope_name + '_conv2')

        # The identity path
        n_input_channels = inputs.shape.as_list()[3]
        if n_input_channels == n_fmaps:
            path2 = inputs
        else:
            path2 = conv2d(inputs, [1, 1], [1] + stride + [1], n_fmaps, FLAGS.w_std, FLAGS.b_init, True, 'SAME',
                           scope_name + '_transform')

        # Add and return
        assert path1.shape.as_list() == path2.shape.as_list()
        out = path1 + path2
        out = batch_norm(istrain, out, scope_name + '_addition')
        out = tf.nn.relu(out)
        return out

    def resnet_block(inputs, noiseemb, cleanemb, kernel_size, stride, n_fmaps, scope_name):
        """
        Residual block to process noisy signals, with injection of embedding vectors
        :param inputs      : input feature maps
        :param noiseposemb : positive embedding vector
        :param noisenegemb : negative embedding vector
        :param n_fmaps     : number of channels
        :return out        : output feature maps
        """

        def cont_embed(n, out_dim, scope_name):
            out = tf.constant(list(range(0, n)), dtype=tf.float32)  # [n]
            out = tf.reshape(out, [n, 1])  # [n, 1]
            out = dense(out, 50, FLAGS.w_std, 0.0, False, scope_name + '_dense1')  # [n, 50]
            out = batch_norm(istrain, out, scope_name + scope_name + '_dense1')
            out = tf.nn.relu(out)
            out = dense(out, 50, FLAGS.w_std, 0.0, False, scope_name + '_dense2')  # [n, 50]
            out = batch_norm(istrain, out, scope_name + scope_name + '_dense2')
            out = tf.nn.relu(out)
            out = dense(out, out_dim, 0.0, 0.0, False, scope_name + '_dense3')  # [n, out_dim]
            return out

        def process_noise_t_f(match_to, scope_name):
            n_fmaps = match_to.shape[3].value
            # Project the noise to fit the conv
            noise_proj = dense(noiseemb, n_fmaps, 0.0, 0.0, True, scope_name + '_noise_emb')  # [mb, n_fmaps]
            noise_proj = tf.expand_dims(noise_proj, 1)
            noise_proj = tf.expand_dims(noise_proj, 1)  # [mb, 1, 1, n_fmaps]

            clean_proj = dense(cleanemb, n_fmaps, 0.0, 0.0, True, scope_name + '_clean_emb')  # [mb, n_fmaps]
            clean_proj = tf.expand_dims(clean_proj, 1)
            clean_proj = tf.expand_dims(clean_proj, 1)  # [mb, 1, 1, n_fmaps]

            # Get the time and frequency embedding
            ts, fs = match_to.shape[1].value, match_to.shape[2].value
            tout = cont_embed(ts, n_fmaps, scope_name + '_temb')  # [ts, n_fmaps]
            tout = tf.expand_dims(tout, 1)
            tout = tf.expand_dims(tout, 0)  # [1, time, 1, n_fmaps]
            fout = cont_embed(fs, n_fmaps, scope_name + '_femb')  # [fs, n_fmaps]
            fout = tf.expand_dims(fout, 0)
            fout = tf.expand_dims(fout, 0)  # [1, 1, freq, n_fmaps]

            return noise_proj, clean_proj, tout, fout

        # The transformation path
        path1 = conv2d(inputs, [kernel_size, kernel_size], [1, stride, stride, 1],
                       n_fmaps, FLAGS.w_std, FLAGS.b_init, False,
                       'SAME', scope_name + '_conv1')  # [mb, time, freq, n_fmaps]
        noise_proj1, clean_proj1, tout1, fout1 = process_noise_t_f(path1, scope_name + '_conv1')
        path1 = path1 + noise_proj1 + clean_proj1 + tout1 + fout1
        path1 = batch_norm(istrain, path1, scope_name + '_conv1')
        path1 = tf.nn.relu(path1)
        path1 = conv2d(path1, [kernel_size, kernel_size], [1, 1, 1, 1], n_fmaps, FLAGS.w_std, FLAGS.b_init, True,
                       'SAME', scope_name + '_conv2')
        noise_proj2, clean_proj2, tout2, fout2 = process_noise_t_f(path1, scope_name + '_conv2')
        path1 = path1 + noise_proj2 + clean_proj2 + tout2 + fout2

        # The identity path
        n_input_channels = inputs.shape.as_list()[3]
        if n_input_channels == n_fmaps:
            path2 = inputs
        else:
            path2 = conv2d(inputs, [1, 1], [1, stride, stride, 1], n_fmaps, FLAGS.w_std, FLAGS.b_init, True, 'SAME',
                           scope_name + '_transform')

        # Add and return
        assert path1.shape.as_list() == path2.shape.as_list()
        out = path1 + path2
        out = batch_norm(istrain, out, scope_name + '_addition')
        out = tf.nn.relu(out)
        return out

    # The interference speaker embedding
    with tf.compat.v1.variable_scope('embedding'):
        nout = noisecontext  # [mb, noise frames, 201]
        nout = tf.expand_dims(nout, 3)
        nout = noise_resnet_block(nout, [8, 4], [3, 2], 64, 'noise_resblock1_1')  # [mb, noise frames, 201, 64]
        nout = noise_resnet_block(nout, [8, 4], [3, 2], 128, 'noise_resblock2_1')  # [mb, noise frames / 2, 201 / 2, 64]
        nout = noise_resnet_block(nout, [4, 4], [1, 1], 256, 'noise_resblock3_1')  # [mb, noise frames / 4, 201 / 4, 64]
        nout = noise_resnet_block(nout, [4, 4], [1, 2], 512,
                                  'noise_resblock4_1')  # [mb, noise frames / 8, 201 / 8, 512]
        nout = tf.nn.avg_pool2d(nout, [1, nout.shape[1].value, nout.shape[2].value, 1], [1, 1, 1, 1],
                                'VALID')  # [mb, 1, 1, 512]
        assert nout.shape.as_list()[1:3] == [1, 1]
        noiseemb = nout[:, 0, 0, :]  # [mb, 512]

    # The target speaker embedding
    with tf.compat.v1.variable_scope('embedding', reuse=True):
        nout = cleancontext  # [mb, noise frames, 201]
        nout = tf.expand_dims(nout, 3)
        nout = noise_resnet_block(nout, [8, 4], [3, 2], 64, 'noise_resblock1_1')  # [mb, noise frames, 201, 64]
        nout = noise_resnet_block(nout, [8, 4], [3, 2], 128, 'noise_resblock2_1')  # [mb, noise frames / 2, 201 / 2, 64]
        nout = noise_resnet_block(nout, [4, 4], [1, 1], 256, 'noise_resblock3_1')  # [mb, noise frames / 4, 201 / 4, 64]
        nout = noise_resnet_block(nout, [4, 4], [1, 2], 512,
                                  'noise_resblock4_1')  # [mb, noise frames / 8, 201 / 8, 512]
        nout = tf.nn.avg_pool2d(nout, [1, nout.shape[1].value, nout.shape[2].value, 1], [1, 1, 1, 1],
                                'VALID')  # [mb, 1, 1, 512]
        assert nout.shape.as_list()[1:3] == [1, 1]
        cleanemb = nout[:, 0, 0, :]  # [mb, 512]

    # Processing the mixed signal
    out = mixed  # [mb, context frames, 201]
    out = tf.expand_dims(out, 3)
    out = resnet_block(out, noiseemb, cleanemb, 4, 1, 64, 'resblock1_1')
    out = resnet_block(out, noiseemb, cleanemb, 4, 1, 64, 'resblock1_2')
    out = resnet_block(out, noiseemb, cleanemb, 4, 2, 128, 'resblock2_1')
    out = resnet_block(out, noiseemb, cleanemb, 4, 1, 128, 'resblock2_2')
    out = resnet_block(out, noiseemb, cleanemb, 3, 2, 256, 'resblock3_1')
    out = resnet_block(out, noiseemb, cleanemb, 3, 1, 256, 'resblock3_2')
    out = resnet_block(out, noiseemb, cleanemb, 3, 2, 512, 'resblock4_1')
    out = resnet_block(out, noiseemb, cleanemb, 3, 1, 512, 'resblock4_2')  # [mb, context frames / 8, 201 / 8, 512]

    # final layers
    out = conv2d(out, [out.shape[1].value, 1], [1, 1, 1, 1],
                 512, FLAGS.w_std, FLAGS.b_init, False,
                 'VALID', 'last_conv')  # [mb, 1, 201 / 8, 512]
    out = batch_norm(istrain, out, 'last_conv')
    out = tf.nn.relu(out)
    out = flatten(out)  # [mb,  (201 / 8) * 512]
    out = dense(out, nfeat, 0.0, 0.0, True, 'last_dense')  # [mb, 201]
    mixed_central = mixed[:, FLAGS.window_frames // 2, :]  # [mb, 201]
    denoised = mixed_central + out  # [mb, 201]

    # Loss
    se = tf.square(denoised - clean[:, 0, :])  # [mb, 201]
    imp_factor = np.linspace(2, 1, nfeat, dtype=np.float32).reshape((1, nfeat))
    example_loss = tf.reduce_mean(se * tf.constant(imp_factor), axis=1)
    loss = tf.reduce_mean(example_loss)

    monitors = {'loss': loss}
    outputs = {'loss': example_loss, 'mixed': mixed_central, 'denoised': denoised, 'clean': clean[:, 0, :],
               'mixedph': mixedph[:, 0, :], 'location': location, 'cleanpath': cleanpath,
               'noisepath': noisepath, 'snr': snr}
    return loss, monitors, outputs


"""
Restoration from Estimated Spectrum
"""


def evaluate(outputs, ereader, step):
    """
    Restoration from Estimated Spectrum
    :param outputs : estimated spectrograms
    :param ereader : evaluation data reader
    :param step    : global step
    """
    print(ereader.name)

    # Print the loss
    loss = outputs['loss'].mean()
    print('loss: {}'.format(loss))

    # Create a summary
    summary = tf.compat.v1.Summary()
    summary.value.add(tag='{}_{}'.format(ereader.name, 'loss'), simple_value=loss)

    # Construct the tf graph for reconstruction
    g = tf.Graph()
    with g.as_default():
        stft = tf.compat.v1.placeholder(dtype=tf.complex64, shape=[None, 201])
        samplestf = tf.signal.inverse_stft(tf.expand_dims(stft, 0), ereader.frame_length, ereader.frame_step,
                                           ereader.frame_length, window_fn=tf.signal.inverse_stft_window_fn(ereader.frame_step, forward_window_fn=functools.partial(tf.signal.hann_window, periodic=True)))[0, :]
        sess = tf.compat.v1.Session()

    # Reconstruct some audio
    startlocations = np.where(outputs['location'] == 0)[0]
    for i in range(0, len(startlocations)):
        # Start and end location
        s = startlocations[i]
        if i == len(startlocations) - 1:
            e = len(outputs['mixed'])
        else:
            e = startlocations[i + 1]

        # Collect frames
        mixedmagframes = outputs['mixed'][s:e]
        mixedphframes = outputs['mixedph'][s:e]
        denoisedmagframes = outputs['denoised'][s:e]

        # Undo the log operation
        mixedmagframes = np.exp(mixedmagframes)
        denoisedmagframes = np.exp(denoisedmagframes)

        # Reconstruct complex spectrogram
        mixedspect = mixedmagframes * (np.exp(1j * mixedphframes))
        denoisedspect = denoisedmagframes * (np.exp(1j * mixedphframes))

        # Reconstruct waveform
        mixedsamples = sess.run(samplestf, feed_dict={stft: mixedspect})
        denoisedsamples = sess.run(samplestf, feed_dict={stft: denoisedspect})

        # Save the waveforms
        cleanpath, noisepath, snr = outputs['cleanpath'][s], outputs['noisepath'][s], outputs['snr'][s]
        cleanpath, noisepath = cleanpath.decode('utf-8'), noisepath.decode('utf-8')
        cleanpath = cleanpath.split('/')[-1][:-4]
        noisepath = noisepath.split('/')[-1][:-4]
        mixed_filename = '{}_{}_{}_{}_{}_{}.wav'.format(modelname, step, cleanpath, noisepath, snr, 'mixed')
        denoised_filename = '{}_{}_{}_{}_{}_{}.wav'.format(modelname, step, cleanpath, noisepath, snr, 'denoised')
        scipy.io.wavfile.write(os.path.join(FLAGS.wav_dump_folder, mixed_filename), 16000, mixedsamples)
        scipy.io.wavfile.write(os.path.join(FLAGS.wav_dump_folder, denoised_filename), 16000, denoisedsamples)


"""
Main Loop for Training and Evaluation with Monitors
"""


def main_loop():
    global treader, tsess, tph, modelname, isess, tin

    # Print FLAGS
    _ = FLAGS.lr
    print('----------------------------- FLAGS VALUES --------------------------------')
    for k in sorted(FLAGS.__flags.keys()):
        print('{}: {}'.format(k, getattr(FLAGS, k)))
    print('')

    # Print General model information
    print('----------------------- DATA LOADING, MODEL PREPARING -------------------------')
    time_stamp = str(datetime.now()).replace(':', '-').replace(' ', '_')
    modelname = FLAGS.model_name if FLAGS.model_name else str(time_stamp)
    print('model_name: {}'.format(modelname))

    ig = tf.Graph()
    with ig.as_default():
        with tf.device('/cpu:0'):
            # Preparing the reader
            treader = get_train_reader()
            treader.preparations()

            # Readers
            ereaders = get_eval_readers()

        isess = tf.compat.v1.Session(graph=ig, config=tf.compat.v1.ConfigProto(log_device_placement=False,
                                                                               allow_soft_placement=True))

    # Training model settings
    tg = tf.Graph()
    with tg.as_default():
        # Session
        tsess = tf.compat.v1.Session(graph=tg, config=tf.compat.v1.ConfigProto(log_device_placement=False,
                                                                               allow_soft_placement=True))

        # The model
        with tf.device('/gpu:0'):
            tin = treader.get_inputs()
            loss, monitors, _ = model(tin, True)
            # The global step
            global_step = tf.Variable(0, trainable=False)
            # Optimization
            if FLAGS.alg == 'sgd':
                opt = tf.compat.v1.train.GradientDescentOptimizer(FLAGS.lr)
            elif FLAGS.alg == 'momentum':
                opt = tf.train.MomentumOptimizer(FLAGS.lr, FLAGS.mom)
            elif FLAGS.alg == 'rmsprop':
                opt = tf.train.RMSPropOptimizer(FLAGS.lr, momentum=FLAGS.mom)
            elif FLAGS.alg == 'adadelta':
                opt = tf.train.AdadeltaOptimizer(FLAGS.lr)
            elif FLAGS.alg == 'adagrad':
                opt = tf.train.AdagradOptimizer(FLAGS.lr)
            elif FLAGS.alg == 'adam':
                opt = tf.train.AdamOptimizer(learning_rate=FLAGS.lr)
            grads_and_vars = opt.compute_gradients(loss, tf.compat.v1.trainable_variables())
            train_op = opt.apply_gradients(grads_and_vars)
            print('#trainable variables: {}'.format(
                sum([x.get_shape().num_elements() for x in tf.compat.v1.trainable_variables()])))
            print('#non-trainable variables: {}'.format(sum([x.get_shape().num_elements() for x in
                                                             set(tf.compat.v1.global_variables()) - set(
                                                                 tf.compat.v1.trainable_variables())])))
            assert not tf.compat.v1.local_variables()  # Should be empty always, otherwise change the lines above

        # Save, restore and init
        tsaver = tf.compat.v1.train.Saver(tf.compat.v1.global_variables(), max_to_keep=FLAGS.checkpoints_to_keep)
        if FLAGS.restore_path:
            # Restore the variables values
            print('Restoring model from {}'.format(FLAGS.restore_path))
            restore_saver = tf.compat.v1.train.Saver(tf.compat.v1.global_variables())
            restore_saver.restore(tsess, FLAGS.restore_path)
        else:
            tsess.run(tf.compat.v1.global_variables_initializer())

    # Evaluation model settings
    eg = tf.Graph()
    with eg.as_default():
        esess = tf.compat.v1.Session(config=tf.compat.v1.ConfigProto(allow_soft_placement=True))
        with tf.device('/gpu:0'):
            ein = ereaders[0].get_inputs()
            _, _, outputs = model(ein, False)
        esaver = tf.compat.v1.train.Saver(tf.compat.v1.global_variables())

    # Training loop
    print('--------------------------------- TRAINING! ------------------------------------')

    # Init aggregators and names
    aggregators = {k: np.zeros(shape=monitors[k].shape.as_list()) for k in monitors.keys()}

    # Operations that need to be executed in every training step
    step_ops = {'train_op': train_op}
    step_ops.update(monitors)

    # Measure time
    start = time.time()

    # The step
    tstep = 0

    try:
        # Start queue runners
        with ig.as_default():
            with tf.device('/cpu:0'):
                isess.run(tf.compat.v1.local_variables_initializer())
                icoord = tf.train.Coordinator()
                ithreads = tf.train.start_queue_runners(sess=isess, coord=icoord, start=False)
                queue_stats = tf.stack(
                    [x.queue.size() for x in tf.compat.v1.get_collection(tf.compat.v1.GraphKeys.QUEUE_RUNNERS)])
                for t in ithreads:
                    t.start()

        def save_and_eval():
            # Saving and restoring
            print('Saving and restoring the model')
            tsaver.save(tsess,
                        os.path.join(FLAGS.checkpoint_dir, modelname),
                        global_step=tstep,
                        latest_filename=modelname)
            ckpt = tf.train.get_checkpoint_state(FLAGS.checkpoint_dir, latest_filename=modelname)
            if ckpt and ckpt.model_checkpoint_path:
                print('Restoring from file: {}'.format(ckpt.model_checkpoint_path))
                global_step = ckpt.model_checkpoint_path.split('/')[-1].split('-')[-1]
            else:
                print('No checkpoint file found')
                return
            esaver.restore(esess, ckpt.model_checkpoint_path)

            # Print the header
            print('----------------- TEST MONITOR ----------------------')

            for ereader in ereaders:
                # Initialize aggregators
                aggregators = {k: [] for k in outputs.keys()}

                # Build the graph again
                try:
                    with ig.as_default():
                        with tf.device('/cpu:0'):
                            ereader.preparations()
                            isess.run(tf.compat.v1.local_variables_initializer())
                            nowcoord = tf.train.Coordinator()
                            nowthreads = []
                            qr1 = tf.compat.v1.get_collection(tf.compat.v1.GraphKeys.QUEUE_RUNNERS)[-3]
                            qr2 = tf.compat.v1.get_collection(tf.compat.v1.GraphKeys.QUEUE_RUNNERS)[-2]
                            qr3 = tf.compat.v1.get_collection(tf.compat.v1.GraphKeys.QUEUE_RUNNERS)[-1]
                            nowthreads.extend(qr1.create_threads(isess, coord=nowcoord, daemon=True, start=True))
                            nowthreads.extend(qr2.create_threads(isess, coord=nowcoord, daemon=True, start=True))
                            nowthreads.extend(qr3.create_threads(isess, coord=nowcoord, daemon=True, start=True))

                    i = 0
                    while True:
                        i += 1

                        input_feed_dict = {ereader.mb: FLAGS.eval_mb}
                        try:
                            batch = isess.run(ereader.batch, feed_dict=input_feed_dict)
                        except tf.errors.OutOfRangeError:
                            break
                        QUICK_MODE = True
                        if QUICK_MODE and i == 2:
                            break
                        model_feed_dict = {a: b for (a, b) in zip(ein, batch)}
                        results = esess.run(outputs, model_feed_dict)

                        # concatenation aggregation of certain channels
                        for k in results.keys():
                            aggregators[k].append(results[k])
                finally:
                    nowcoord.request_stop()
                    nowcoord.join(nowthreads)

                # Concatenate results over different batches
                for k in results.keys():
                    aggregators[k] = np.concatenate(aggregators[k])

                # Dump activations
                if FLAGS.dump_results:
                    for k in results.keys():
                        np.save(join(FLAGS.dump_results, '{}_{}_{}_{}'.format(modelname, ereader.name, tstep, k)),
                                aggregators[k])

                # Evaluate
                evaluate(aggregators, ereader, global_step)

            # Print the bottom part
            print('-----------------------------------------------------')
            print('')

        # Eval before training
        if FLAGS.eval_before_training:
            save_and_eval()

        # Training loop
        while tstep < FLAGS.batches and not icoord.should_stop():
            # Running the step
            input_feed_dict = {treader.mb: FLAGS.train_mb}
            # t1 = time.time()
            batch = isess.run(treader.batch, feed_dict=input_feed_dict)
            # t2 = time.time()
            model_feed_dict = {a: b for (a, b) in zip(tin, batch)}
            results = tsess.run(step_ops, model_feed_dict)

            # Aggregate the train monitoring channels
            for key in monitors.keys():
                aggregators[key] += results[key]

            # Increase the step. 'step' always contains the number of gradient updates that were performed.
            tstep += 1

            # Train monitoring
            if tstep % FLAGS.train_monitor_every == 0:
                # Do some printing
                print('----- TRAIN MONITOR AFTER ANOTHER {} BATCHES ------------'.format(FLAGS.train_monitor_every))
                print('step number: {}'.format(tstep))
                for k in sorted(aggregators.keys()):
                    aggregators[k] = aggregators[k] / float(FLAGS.train_monitor_every)
                    print('{}: {}'.format(k, aggregators[k]))  # loss
                end = time.time()
                print('queues stats: {}'.format(isess.run(queue_stats)))
                print('seconds elapsed: {}'.format(end - start))
                start = end
                # Add summaries and reset aggregators
                summary = tf.compat.v1.Summary()
                summary.value.add(tag='{}_{}'.format(treader.name, 'loss'), simple_value=aggregators['loss'])
                # treader.swriter.add_summary(summary, tstep)
                aggregators = {k: 0.0 for k in monitors.keys()}
                print('---------------------------------------------------------')
                print('')

            # Save and eval
            if tstep % FLAGS.eval_every == 0:
                save_and_eval()

        # Eval after training
        if FLAGS.eval_after_training:
            save_and_eval()

    finally:
        icoord.request_stop()
        time.sleep(1)
        isess.close()
        tsess.close()
        esess.close()
        icoord.join(ithreads)
  
      
if __name__ == '__main__':
    main_loop()



import codecs
import os
import sys

try:
	from setuptools import setup
except:
	from distutils.core import setup


def read(fname):
	return codecs.open(os.path.join(os.path.dirname(__file__), fname)).read()


classifiers = [
    'Development Status :: 4 - Beta',
    'Topic :: Multimedia :: Sound/Audio',
    'Topic :: Multimedia :: Sound/Audio :: Speech',
    'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
    'Programming Language :: Python :: 3',
]

install_requires = [
    'numpy>=1.14.5',
    'tensorflow_gpu>=1.14.0',
    'scipy>=1.0.1'
    'sox>=1.3.7'
]

    
setup(name='N-HANS',
      version='2020.06.22',
      description='Neuro-Holistic Audio-eNhancement System',
      long_description='',
	  packages=['N_HANS','N_HANS.N_HANS___Selective_Noise','N_HANS.N_HANS___Source_Separation'],
      author="Shuo Liu, Gil Keren, Bjoern Schuller",
      author_email="shuo.liu@informatik.uni-augsburg.de",
      url="https://github.com/N-HANS/N-HANS",
      license="License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
      # platforms=['linux'],
      classifiers=classifiers,
      install_requires=install_requires,
	  scripts = [],
	  entry_points = {
		'console_scripts': [
			'nhans_denoiser = N_HANS.N_HANS___Selective_Noise.apply:main',
			'load_denoiser = N_HANS.N_HANS___Selective_Noise.load_model:main',
			'nhans_separator = N_HANS.N_HANS___Source_Separation.apply:main',
			'load_separator = N_HANS.N_HANS___Source_Separation.load_model:main',
		]
	  }
)



########################################################################################################################
#                                          N-HANS speech separator                                                     #
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
#   Description:      Loading separator model.                                                                         #
#   Authors:          Shuo Liu, Gil Keren, Bjoern Schuller                                                             #
#   Affiliation:      Chair of Embedded Intelligence for Health Care and Wellbeing, University of Augsburg (UAU)  #
#   Version:          2.0                                                                                              #
#   Last Update:      May. 06, 2020                                                                                    #
#   Dependence Files: xxx                                                                                              #
#   Contact:          shuo.liu@informatik.uni-augburg.de                                                               #
########################################################################################################################
#                                                                                                                      #
#   Copyright (C) 2019, Shuo Liu, Gil Keren and Björn Schuller: University of Augsburg.                                #
#                                                                                                                      #
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public    #
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any       #
# later version.                                                                                                       #
#                                                                                                                      #
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied   #
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.#
#                                                                                                                      #
# You should have received a copy of the GNU General Public License along with this program. #  If not, see            #
# <http://www.gnu.org/licenses/>.                                                                                      #
#                                                                                                                      #
########################################################################################################################

import os


def main():
    dropbox_link = 'https://www.dropbox.com/s/wdk62o4kt8lbvjr/N_HANS___Separator.zip?dl=0'
    download_filename = 'N_HANS_separator.zip'
    command_download = 'wget {} -O {}'.format(dropbox_link, download_filename)
    command_unzip = 'unzip N_HANS_separator.zip'

    os.system(command_download)
    os.system(command_unzip)

